# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from itertools import islice

from gluonts.dataset.artificial import constant_dataset
from gluonts.mx import DeepAREstimator
from gluonts.mx.distribution import StudentTOutput
from gluonts.mx.trainer import Trainer
from gluonts.mx.util import get_hybrid_forward_input_names

ds_info, train_ds, test_ds = constant_dataset()
freq = ds_info.metadata.freq
prediction_length = ds_info.prediction_length


def test_distribution():
    """
    Makes sure additional tensors can be accessed and have expected shapes
    """
    prediction_length = ds_info.prediction_length

    # todo adapt loader to anomaly detection use-case
    batch_size = 2
    num_samples = 3

    estimator = DeepAREstimator(
        freq=freq,
        prediction_length=prediction_length,
        trainer=Trainer(epochs=2, num_batches_per_epoch=1),
        distr_output=StudentTOutput(),
        batch_size=batch_size,
    )

    train_output = estimator.train_model(train_ds, test_ds)

    training_data_loader = estimator.create_training_data_loader(
        estimator.create_transformation().apply(train_ds)
    )

    seq_len = 2 * ds_info.prediction_length

    for data_entry in islice(training_data_loader, 1):
        input_names = get_hybrid_forward_input_names(
            type(train_output.trained_net)
        )

        distr = train_output.trained_net.distribution(
            *[data_entry[k] for k in input_names]
        )

        assert distr.sample(num_samples).shape == (
            num_samples,
            batch_size,
            seq_len,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List

import pytest
import torch

from gluonts.torch.util import (
    lagged_sequence_values,
    unsqueeze_expand,
)


@pytest.mark.parametrize(
    "lag_indices, prior_sequence, sequence, output_shape",
    [
        (
            [0, 1, 5, 10, 20],
            torch.randn((4, 100)),
            torch.randn((4, 8)),
            (4, 8, 5),
        ),
        (
            [0, 1, 5, 10, 20],
            torch.randn((4, 100, 1)),
            torch.randn((4, 8, 1)),
            (4, 8, 1, 5),
        ),
        (
            [0, 1, 5, 10, 20],
            torch.randn((4, 100, 2)),
            torch.randn((4, 8, 2)),
            (4, 8, 2, 5),
        ),
    ],
)
def test_lagged_sequence_values(
    lag_indices: List[int],
    prior_sequence: torch.Tensor,
    sequence: torch.Tensor,
    output_shape: tuple,
):
    res = lagged_sequence_values(lag_indices, prior_sequence, sequence, dim=1)
    assert res.shape == output_shape
    full_sequence = torch.cat((prior_sequence, sequence), dim=1)
    for t in range(res.shape[1]):
        expected_lags_t = torch.stack(
            [
                full_sequence[:, t + prior_sequence.shape[1] - l]
                for l in lag_indices
            ],
            dim=-1,
        )
        assert expected_lags_t.shape == res[:, t, :].shape
        assert torch.allclose(expected_lags_t, res[:, t, :])


@pytest.mark.parametrize(
    "input_shape, dim, size, output_shape",
    [
        ((3, 4, 5), 2, 42, (3, 4, 42, 5)),
        ((3, 4, 5), 0, 42, (42, 3, 4, 5)),
        ((3, 4, 5), -1, 42, (3, 4, 5, 42)),
        ((3, 4, 5), -2, 42, (3, 4, 42, 5)),
    ],
)
def test_unsqueeze_expand(input_shape, dim, size, output_shape):
    a = torch.randn(size=input_shape)
    b = unsqueeze_expand(a, dim=dim, size=size)
    assert b.shape == output_shape



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from gluonts.core.component import validated
from gluonts.dataset.common import Dataset
from gluonts.model.estimator import Estimator, Predictor

from ._predictor import TreePredictor


class ThirdPartyEstimator(Estimator):
    """
    An `Estimator` that uses an external fitting mechanism, thus eliminating
    the need for a Trainer. Differs from DummyEstimator in that DummyEstimator
    does not use the training data, but merely trains at prediction time.

    Parameters
    ----------
    predictor_cls
        `Predictor` class to instantiate.
    **kwargs
        Keyword arguments to pass to the predictor constructor.
    """

    @validated()
    def __init__(self, predictor_cls: type, **kwargs) -> None:
        self.predictor = predictor_cls(**kwargs)

    def train(
        self, training_data: Dataset, validation_dataset=None
    ) -> Predictor:
        return self.predictor.train(training_data)


class TreeEstimator(ThirdPartyEstimator):
    @validated(
        getattr(TreePredictor.__init__, "Model")
    )  # Reuse the model Predictor model
    def __init__(self, **kwargs) -> None:
        super().__init__(predictor_cls=TreePredictor, **kwargs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass
from typing import Callable, ChainMap, Dict

import numpy as np

from .aggregations import Aggregation


@dataclass
class Evaluator:
    name: str

    def update(self, data: ChainMap[str, np.ndarray]) -> None:
        raise NotImplementedError

    def get(self) -> np.ndarray:
        raise NotImplementedError


@dataclass
class DirectEvaluator(Evaluator):
    """An Evaluator which uses a single function and aggregation strategy."""

    stat: Callable
    aggregate: Aggregation

    def update(self, data: ChainMap[str, np.ndarray]) -> None:
        self.aggregate.step(self.stat(data))

    def get(self) -> np.ndarray:
        return self.aggregate.get()


@dataclass
class DerivedEvaluator(Evaluator):
    """An Evaluator for metrics that are derived from other metrics.

    A derived metric updates multiple, simpler metrics independently and in
    the end combines their results as defined in `post_process`."""

    evaluators: Dict[str, Evaluator]
    post_process: Callable

    def update(self, data: ChainMap[str, np.ndarray]) -> None:
        for evaluator in self.evaluators.values():
            evaluator.update(data)

    def get(self) -> np.ndarray:
        return self.post_process(
            **{
                name: evaluator.get()
                for name, evaluator in self.evaluators.items()
            }
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from pathlib import Path

import numpy as np
import pandas as pd

from gluonts.ext.naive_2 import naive_2

# DISCLAIMER:
# this script is only to test whether the R implementation
# of naive_2 produces the same outputs as out python implementation

R_INPUT_FILE = "r_naive_2_inputs.csv"
R_OUTPUT_FILE = "r_naive_2_outputs.csv"


def load_naive_2_data():
    test_directory_path = Path(os.getenv("PYTEST_CURRENT_TEST")).parents[0]
    r_naive_2_inputs = pd.read_csv(test_directory_path / R_INPUT_FILE).values
    r_naive_2_outputs = pd.read_csv(test_directory_path / R_OUTPUT_FILE).values
    return r_naive_2_inputs, r_naive_2_outputs


# To generate the above dataset use the following script:
# https://github.com/Mcompetitions/M4-methods/blob/master/Benchmarks%20and%20Evaluation.R
# and then paste the following code in as the first line:
"""
set.seed(1234567)
"""
# an the below code at the very end:
"""
for (i in 1:length(data_train)){
   insample <- data_train[[i]]
   d <- insample[1:14]
   v <-paste(d, collapse=", ")
   cat(c("[",v,"],"))
}

print('\n')

for (i in 1:length(data_train)){
  
  insample <- data_train[[i]]
  forecasts <- Benchmarks(input=insample, fh=fh)

  f <- forecasts[[1]][1:6]
  v <-paste(f, collapse=", ")
  cat(c("[",v,"],"))

}
"""

# R script variables:
FH = 6  # The forecasting horizon examined
FRQ = 1  # The frequency of the data


def test_naive_2(prediction_length=FH, season_length=FRQ):
    r_naive_2_inputs, r_naive_2_outputs = load_naive_2_data()
    predictions = []
    for i in range(len(r_naive_2_inputs)):
        predictions.append(
            naive_2(
                r_naive_2_inputs[i],
                prediction_length=prediction_length,
                season_length=season_length,
            )
        )
    predictions = np.array(predictions)

    assert np.allclose(r_naive_2_outputs, predictions)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pytest

from gluonts.evaluation.metrics import (
    abs_error,
    abs_target_mean,
    abs_target_sum,
    calculate_seasonal_error,
    coverage,
    mape,
    mase,
    mse,
    msis,
    quantile_loss,
    smape,
)

ZEROES = np.array([0.0] * 5)
LINEAR = np.array([0.0, 0.1, 0.2, 0.3, 0.4])
EXPONENTIAL = np.array([0.1, 0.01, 0.001, 0.0001, 0.00001])
CONSTANT = np.array([0.4] * 5)


# TODO remove `"exclude_zero_denominator": True` when metrics are refactored
@pytest.mark.parametrize(
    "target, forecast, metrics",
    [
        (
            ZEROES,
            ZEROES,
            [
                (mase, np.nan, {"seasonal_error": 0.0}),
                (mase, 0.0, {"seasonal_error": 1.0}),
                (mse, 0.0, {}),
                (abs_error, 0.0, {}),
                (quantile_loss, 0.0, {"q": 0.5}),
                (coverage, 1.0, {}),
                (mape, np.nan, {}),
                (smape, np.nan, {}),
            ],
        ),
        (
            LINEAR,
            ZEROES,
            [
                (mase, 0.2 / 10e-10, {"seasonal_error": 10e-10}),
                (mase, np.inf, {"seasonal_error": 0.0}),
                (mse, 0.06, {}),
                (abs_error, 1.0, {}),
                (quantile_loss, 0.2, {"q": 0.1}),
                (quantile_loss, 1.0, {"q": 0.5}),
                (quantile_loss, 1.8, {"q": 0.9}),
                (coverage, 0.2, {}),
                (mape, np.nan, {}),
                (smape, np.nan, {}),
            ],
        ),
        (
            LINEAR,
            CONSTANT,
            [
                (mase, 0.2, {"seasonal_error": 1}),
                (mse, 0.06, {}),
                (abs_error, 1.0, {}),
                (quantile_loss, 1.8, {"q": 0.1}),
                (quantile_loss, 1.0, {"q": 0.5}),
                (quantile_loss, 0.2, {"q": 0.9}),
                (coverage, 1.0, {}),
                (mape, np.inf, {}),
                (smape, 0.8304761904761906, {}),
            ],
        ),
        (
            ZEROES,
            EXPONENTIAL,
            [
                (mase, 0.022222, {"seasonal_error": 1}),
                (mse, 0.00202020202, {}),
                (abs_error, 0.11111, {}),
                (quantile_loss, 0.199998, {"q": 0.1}),
                (quantile_loss, 0.11111, {"q": 0.5}),
                (quantile_loss, 0.0222219, {"q": 0.9}),
                (coverage, 1.0, {}),
                (mape, np.inf, {}),
                (smape, 2.0, {}),
            ],
        ),
        (
            np.array([1.0, 2.0, 3.0, 0.0, 0.0]),
            np.array([2.0, 3.0, 4.0, 1.0, 0.0]),
            [
                (
                    mase,
                    0.8,
                    {"seasonal_error": 1},
                ),
                (mse, 0.8, {}),
                (abs_error, 4.0, {}),
                (quantile_loss, 7.2, {"q": 0.1}),
                (quantile_loss, 4.0, {"q": 0.5}),
                (quantile_loss, 0.7999999, {"q": 0.9}),
                (coverage, 1.0, {}),
                (mape, np.nan, {}),
                (smape, np.nan, {}),
            ],
        ),
        (
            CONSTANT,
            EXPONENTIAL,
            [
                (
                    mase,
                    0.377778,
                    {"seasonal_error": 1},
                ),
                (mse, 0.14424260202, {}),
                (abs_error, 1.88889, {}),
                (quantile_loss, 0.377778, {"q": 0.1}),
                (quantile_loss, 1.88889, {"q": 0.5}),
                (quantile_loss, 3.400002, {"q": 0.9}),
                (coverage, 0.0, {}),
                (mape, 0.944445, {}),
                (smape, 1.8182728, {}),
            ],
        ),
    ],
)
def test_metrics(target, forecast, metrics):
    for metric, expected, kwargs in metrics:
        np.testing.assert_almost_equal(
            metric(target, forecast, **kwargs), expected
        )


@pytest.mark.parametrize(
    "target, metric, expected",
    [
        (ZEROES, abs_target_sum, 0.0),
        (LINEAR, abs_target_sum, 1.0),
        (ZEROES, abs_target_mean, 0.0),
        (LINEAR, abs_target_mean, 0.2),
    ],
)
def test_target_metrics(target, metric, expected):
    np.testing.assert_almost_equal(metric(target), expected)


@pytest.mark.parametrize(
    "target, lower_quantile, upper_quantile, seasonal_error, alpha, expected",
    [
        (LINEAR, ZEROES, CONSTANT, 0.0, 0.05, np.inf),
        (LINEAR, ZEROES, CONSTANT, 1.0, 0.05, 0.4),
        (LINEAR, ZEROES, CONSTANT, 0.01, 0.05, 40.0),
        (ZEROES, ZEROES, ZEROES, 0.0, 0.05, np.nan),
    ],
)
def test_msis(
    target, lower_quantile, upper_quantile, seasonal_error, alpha, expected
):
    np.testing.assert_almost_equal(
        msis(
            target=target,
            lower_quantile=lower_quantile,
            upper_quantile=upper_quantile,
            seasonal_error=seasonal_error,
            alpha=alpha,
        ),
        expected,
    )


@pytest.mark.parametrize(
    "past_data, seasonality, expected",
    [
        (LINEAR, 1, 0.1),
        (LINEAR, 2, 0.2),
        (LINEAR, 3, 0.3),
        (LINEAR, 4, 0.4),
        (LINEAR, len(LINEAR), 0.1),
        (CONSTANT, 2, 0.0),
        (ZEROES, 1, 0.0),
        (EXPONENTIAL, 3, 0.054945),
    ],
)
def test_seasonal_error(past_data, seasonality, expected):
    np.testing.assert_almost_equal(
        calculate_seasonal_error(
            past_data=past_data, freq=None, seasonality=seasonality
        ),
        expected,
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from copy import deepcopy

import numpy as np
import pandas as pd
import pytest

from gluonts.dataset.common import ListDataset
from gluonts.dataset.field_names import FieldName
from gluonts.transform import (
    AddAgeFeature,
    AddConstFeature,
    AddObservedValuesIndicator,
)

ds1 = [
    {
        "start": pd.Period("2020/01/01", freq="1D"),
        "target": np.array(
            [1, 2, 3, np.nan, 5, np.nan, 7, np.nan, np.nan, 10]
        ),
    }
]
ds2 = ListDataset(
    [
        {
            "start": "2020/01/01",
            "target": [1, 2, 3, np.nan, 5, np.nan, 7, np.nan, np.nan, 10],
        }
    ],
    freq="1D",
)


@pytest.mark.parametrize("ds", [ds1, ds2])
@pytest.mark.parametrize(
    "transform",
    [
        AddObservedValuesIndicator(
            target_field=FieldName.TARGET,
            output_field=FieldName.OBSERVED_VALUES,
        ),
        AddAgeFeature(
            target_field=FieldName.TARGET,
            output_field="age_feature",
            pred_length=1,
        ),
        AddConstFeature(
            target_field=FieldName.TARGET,
            output_field="constant",
            pred_length=1,
        ),
    ],
)
def test_dataset_imutability(ds, transform):
    ds_c = deepcopy(ds)

    # test that using twice the transformation gives the same result
    out1 = transform(ds, is_train=True)
    out2 = transform(ds, is_train=True)
    for o1, o2 in zip(out1, out2):
        for k in o1:
            if isinstance(o1[k], np.ndarray):
                assert np.allclose(o1[k], o2[k], equal_nan=True)
            else:
                assert o1[k] == o2[k]

    # test that the original dataset did not change due to the transformation
    for d, d_c in zip(ds, ds_c):
        for k in d:
            if isinstance(d[k], np.ndarray):
                assert np.allclose(d[k], d_c[k], equal_nan=True)
            else:
                assert d[k] == d_c[k]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx
import numpy as np
import pytest

from gluonts.mx.model.deepvar_hierarchical import (
    constraint_mat,
    coherency_error,
)

TOL = 1e-4

S = np.array(
    [
        [1, 1, 1, 1],
        [1, 1, 0, 0],
        [0, 0, 1, 1],
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ],
)

num_bottom_ts = S.shape[1]
A = constraint_mat(S)


@pytest.mark.parametrize(
    "bottom_ts",
    [
        np.random.randint(low=0, high=100, size=num_bottom_ts),  # integer data
        np.random.randint(
            low=1000, high=100000, size=num_bottom_ts
        ),  # large integer data
        np.random.poisson(lam=1, size=num_bottom_ts),
        np.random.negative_binomial(n=1000, p=0.5, size=num_bottom_ts),
        -np.random.negative_binomial(
            n=1000, p=0.5, size=num_bottom_ts
        ),  # negative data
        np.random.standard_normal(size=num_bottom_ts),
    ],
)
def test_coherency_error(bottom_ts):
    all_ts = S @ bottom_ts

    assert coherency_error(mx.nd.array(A), mx.nd.array(all_ts)) < TOL



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Tuple

import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .distribution import Distribution, _sample_multiple, getF, softplus
from .distribution_output import DistributionOutput


class Laplace(Distribution):
    r"""
    Laplace distribution.

    Parameters
    ----------
    mu
        Tensor containing the means, of shape `(*batch_shape, *event_shape)`.
    b
        Tensor containing the distribution scale, of shape
        `(*batch_shape, *event_shape)`.
    F
    """

    is_reparameterizable = True

    @validated()
    def __init__(self, mu: Tensor, b: Tensor) -> None:
        self.mu = mu
        self.b = b

    @property
    def F(self):
        return getF(self.mu)

    @property
    def batch_shape(self) -> Tuple:
        return self.mu.shape

    @property
    def event_shape(self) -> Tuple:
        return ()

    @property
    def event_dim(self) -> int:
        return 0

    def log_prob(self, x: Tensor) -> Tensor:
        F = self.F
        return -1.0 * (F.log(2.0 * self.b) + F.abs((x - self.mu) / self.b))

    @property
    def mean(self) -> Tensor:
        return self.mu

    @property
    def stddev(self) -> Tensor:
        return 2.0**0.5 * self.b

    def cdf(self, x: Tensor) -> Tensor:
        y = (x - self.mu) / self.b
        return 0.5 + 0.5 * y.sign() * (1.0 - self.F.exp(-y.abs()))

    def sample_rep(self, num_samples=None, dtype=np.float32) -> Tensor:
        F = self.F

        def s(mu: Tensor, b: Tensor) -> Tensor:
            ones = mu.ones_like()
            x = F.random.uniform(-0.5 * ones, 0.5 * ones, dtype=dtype)
            laplace_samples = mu - b * F.sign(x) * F.log(
                (1.0 - 2.0 * F.abs(x)).clip(1.0e-30, 1.0e30)
                # 1.0 - 2.0 * F.abs(x)
            )
            return laplace_samples

        return _sample_multiple(
            s, mu=self.mu, b=self.b, num_samples=num_samples
        )

    def quantile(self, level: Tensor) -> Tensor:
        F = self.F
        for _ in range(self.all_dim):
            level = level.expand_dims(axis=-1)

        condition = F.broadcast_greater(level, level.zeros_like() + 0.5)
        u = F.where(condition, F.log(2.0 * level), -F.log(2.0 - 2.0 * level))

        return F.broadcast_add(self.mu, F.broadcast_mul(self.b, u))

    @property
    def args(self) -> List:
        return [self.mu, self.b]


class LaplaceOutput(DistributionOutput):
    args_dim: Dict[str, int] = {"mu": 1, "b": 1}
    distr_cls: type = Laplace

    @classmethod
    def domain_map(cls, F, mu, b):
        b = F.maximum(softplus(F, b), cls.eps())
        return mu.squeeze(axis=-1), b.squeeze(axis=-1)

    @property
    def event_shape(self) -> Tuple:
        return ()


class LaplaceFixedVarianceOutput(LaplaceOutput):
    @classmethod
    def domain_map(cls, F, mu, b):
        b = 0.01 * F.ones_like(b)
        return mu.squeeze(axis=-1), b.squeeze(axis=-1)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from gluonts.mx import Tensor


class Kernel:
    def kernel_matrix(self, x1: Tensor, x2: Tensor):
        # raise error in the base Kernel class, implement in the concrete
        # subclasses
        raise NotImplementedError()

    def _compute_square_dist(self, F, x1: Tensor, x2: Tensor) -> None:
        r"""
        Parameters
        --------------------
        F : ModuleType
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.
        x1 : Tensor
            Feature data of shape (batch_size, history_length, num_features).
        x2 : Tensor
            Feature data of shape (batch_size, history_length, num_features).

        Returns
        --------------------
        Tensor
            square distance matrix of shape (batch_size, history_length, history_length)
            :math: `\|\mathbf{x_1}-\mathbf{x_2}\|_2^2 = (\mathbf{x_1}-\mathbf{x_2})^T(\mathbf{x_1}-\mathbf{x_2})
                                                        = \|\mathbf{x_1}\|_2^2 - 2\mathbf{x_1}^T\mathbf{x_2}
                                                        + \|\mathbf{x_2}\|_2^2`.
        """  # noqa: E501
        feature_axis = 2
        # Column vector: Add to math:`x_i^Tx_i` to every column in row i
        x1_norm_square = (
            F.norm(x1, ord=2, axis=feature_axis) ** 2
        ).expand_dims(2)
        # Row vector: Add to math:`x_i^Tx_i` to every row in column i
        x2_norm_square = (
            F.norm(x2, ord=2, axis=feature_axis) ** 2
        ).expand_dims(1)
        x1x2_trans = F.linalg.gemm2(x1, x2, transpose_b=True)
        self.square_dist = F.broadcast_add(
            F.broadcast_sub(x1_norm_square, 2 * x1x2_trans), x2_norm_square
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Tuple

import mxnet as mx
import pytest

from gluonts.mx.model.tpp.distribution import Loglogistic, Weibull
from gluonts.mx.distribution import (
    Beta,
    Dirichlet,
    DirichletMultinomial,
    Distribution,
    EmpiricalDistribution,
    Gamma,
    Gaussian,
    Laplace,
    MixtureDistribution,
    MultivariateGaussian,
    NegativeBinomial,
    PiecewiseLinear,
    Poisson,
    StudentT,
    TransformedDistribution,
    Uniform,
)
from gluonts.mx.distribution.bijection import AffineTransformation
from gluonts.mx.distribution.box_cox_transform import BoxCoxTransform
from gluonts.mx.util import make_nd_diag


@pytest.mark.parametrize(
    "distr, expected_batch_shape, expected_event_shape",
    [
        (
            Gaussian(
                mu=mx.nd.zeros(shape=(3, 4, 5)),
                sigma=mx.nd.ones(shape=(3, 4, 5)),
            ),
            (3, 4, 5),
            (),
        ),
        (
            Gamma(
                alpha=mx.nd.ones(shape=(3, 4, 5)),
                beta=mx.nd.ones(shape=(3, 4, 5)),
            ),
            (3, 4, 5),
            (),
        ),
        (
            Beta(
                alpha=mx.nd.ones(shape=(3, 4, 5)),
                beta=mx.nd.ones(shape=(3, 4, 5)),
            ),
            (3, 4, 5),
            (),
        ),
        (
            StudentT(
                mu=mx.nd.zeros(shape=(3, 4, 5)),
                sigma=mx.nd.ones(shape=(3, 4, 5)),
                nu=mx.nd.ones(shape=(3, 4, 5)),
            ),
            (3, 4, 5),
            (),
        ),
        (
            MultivariateGaussian(
                mu=mx.nd.zeros(shape=(3, 4, 5)),
                L=make_nd_diag(F=mx.nd, x=mx.nd.ones(shape=(3, 4, 5)), d=5),
            ),
            (3, 4),
            (5,),
        ),
        (Dirichlet(alpha=mx.nd.ones(shape=(3, 4, 5))), (3, 4), (5,)),
        (
            DirichletMultinomial(
                dim=5, n_trials=9, alpha=mx.nd.ones(shape=(3, 4, 5))
            ),
            (3, 4),
            (5,),
        ),
        (
            Laplace(
                mu=mx.nd.zeros(shape=(3, 4, 5)), b=mx.nd.ones(shape=(3, 4, 5))
            ),
            (3, 4, 5),
            (),
        ),
        (
            NegativeBinomial(
                mu=mx.nd.zeros(shape=(3, 4, 5)),
                alpha=mx.nd.ones(shape=(3, 4, 5)),
            ),
            (3, 4, 5),
            (),
        ),
        (Poisson(rate=mx.nd.ones(shape=(3, 4, 5))), (3, 4, 5), ()),
        (
            Uniform(
                low=-mx.nd.ones(shape=(3, 4, 5)),
                high=mx.nd.ones(shape=(3, 4, 5)),
            ),
            (3, 4, 5),
            (),
        ),
        (
            PiecewiseLinear(
                gamma=mx.nd.ones(shape=(3, 4, 5)),
                slopes=mx.nd.ones(shape=(3, 4, 5, 10)),
                knot_spacings=mx.nd.ones(shape=(3, 4, 5, 10)) / 10,
            ),
            (3, 4, 5),
            (),
        ),
        (
            MixtureDistribution(
                mixture_probs=mx.nd.stack(
                    0.2 * mx.nd.ones(shape=(3, 1, 5)),
                    0.8 * mx.nd.ones(shape=(3, 1, 5)),
                    axis=-1,
                ),
                components=[
                    Gaussian(
                        mu=mx.nd.zeros(shape=(3, 4, 5)),
                        sigma=mx.nd.ones(shape=(3, 4, 5)),
                    ),
                    StudentT(
                        mu=mx.nd.zeros(shape=(3, 4, 5)),
                        sigma=mx.nd.ones(shape=(3, 4, 5)),
                        nu=mx.nd.ones(shape=(3, 4, 5)),
                    ),
                ],
            ),
            (3, 4, 5),
            (),
        ),
        (
            MixtureDistribution(
                mixture_probs=mx.nd.stack(
                    0.2 * mx.nd.ones(shape=(3, 4)),
                    0.8 * mx.nd.ones(shape=(3, 4)),
                    axis=-1,
                ),
                components=[
                    MultivariateGaussian(
                        mu=mx.nd.zeros(shape=(3, 4, 5)),
                        L=make_nd_diag(
                            F=mx.nd, x=mx.nd.ones(shape=(3, 4, 5)), d=5
                        ),
                    ),
                    MultivariateGaussian(
                        mu=mx.nd.zeros(shape=(3, 4, 5)),
                        L=make_nd_diag(
                            F=mx.nd, x=mx.nd.ones(shape=(3, 4, 5)), d=5
                        ),
                    ),
                ],
            ),
            (3, 4),
            (5,),
        ),
        (
            TransformedDistribution(
                StudentT(
                    mu=mx.nd.zeros(shape=(3, 4, 5)),
                    sigma=mx.nd.ones(shape=(3, 4, 5)),
                    nu=mx.nd.ones(shape=(3, 4, 5)),
                ),
                [
                    AffineTransformation(
                        scale=1e-1 + mx.nd.random.uniform(shape=(3, 4, 5))
                    )
                ],
            ),
            (3, 4, 5),
            (),
        ),
        (
            TransformedDistribution(
                MultivariateGaussian(
                    mu=mx.nd.zeros(shape=(3, 4, 5)),
                    L=make_nd_diag(
                        F=mx.nd, x=mx.nd.ones(shape=(3, 4, 5)), d=5
                    ),
                ),
                [
                    AffineTransformation(
                        scale=1e-1 + mx.nd.random.uniform(shape=(3, 4, 5))
                    )
                ],
            ),
            (3, 4),
            (5,),
        ),
        (
            TransformedDistribution(
                Uniform(
                    low=mx.nd.zeros(shape=(3, 4, 5)),
                    high=mx.nd.ones(shape=(3, 4, 5)),
                ),
                [
                    BoxCoxTransform(
                        lambda_1=mx.nd.ones(shape=(3, 4, 5)),
                        lambda_2=mx.nd.zeros(shape=(3, 4, 5)),
                    )
                ],
            ),
            (3, 4, 5),
            (),
        ),
        (
            Loglogistic(
                mx.nd.zeros(shape=(3, 4, 5)),
                mx.nd.ones(shape=(3, 4, 5)),
            ),
            (3, 4, 5),
            (),
        ),
        (
            Weibull(
                mx.nd.ones(shape=(3, 4, 5)),
                mx.nd.ones(shape=(3, 4, 5)),
            ),
            (3, 4, 5),
            (),
        ),
        (
            EmpiricalDistribution(
                samples=mx.nd.random.normal(shape=(10, 3, 4, 5)),
                event_dim=0,
            ),
            (3, 4, 5),
            (),
        ),
        (
            EmpiricalDistribution(
                samples=mx.nd.random.normal(shape=(10, 3, 4, 5)),
                event_dim=1,
            ),
            (3, 4),
            (5,),
        ),
        (
            EmpiricalDistribution(
                samples=mx.nd.random.normal(shape=(10, 3, 4, 5, 20)),
                event_dim=2,
            ),
            (3, 4),
            (5, 20),
        ),
    ],
)
def test_distribution_shapes(
    distr: Distribution,
    expected_batch_shape: Tuple,
    expected_event_shape: Tuple,
):
    assert distr.batch_shape == expected_batch_shape
    assert distr.event_shape == expected_event_shape

    x = distr.sample()

    assert x.shape == distr.batch_shape + distr.event_shape

    loss = distr.loss(x)

    assert loss.shape == distr.batch_shape

    x1 = distr.sample(num_samples=1)

    assert x1.shape == (1,) + distr.batch_shape + distr.event_shape

    x3 = distr.sample(num_samples=3)

    assert x3.shape == (3,) + distr.batch_shape + distr.event_shape

    def has_quantile(d):
        return isinstance(
            d, (Uniform, Gaussian, Laplace, EmpiricalDistribution)
        )

    if (
        has_quantile(distr)
        or isinstance(distr, TransformedDistribution)
        and has_quantile(distr.base_distribution)
    ):
        qs1 = distr.quantile(mx.nd.array([0.5]))
        assert qs1.shape == (1,) + distr.batch_shape + distr.event_shape



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import itertools
import json
import logging
import multiprocessing as mp
import sys
import traceback
from pathlib import Path
from pydoc import locate
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Callable, Iterator, Optional

import numpy as np

import gluonts
from gluonts.core import fqname_for
from gluonts.core.component import equals, from_hyperparameters
from gluonts.core.serde import dump_json, load_json
from gluonts.dataset.common import DataEntry, Dataset
from gluonts.model.forecast import Forecast

if TYPE_CHECKING:  # avoid circular import
    from gluonts.model.estimator import Estimator  # noqa


OutputTransform = Callable[[DataEntry, np.ndarray], np.ndarray]


class Predictor:
    """
    Abstract class representing predictor objects.
    Parameters
    ----------
    prediction_length
        Prediction horizon.
    """

    __version__: str = gluonts.__version__

    def __init__(self, prediction_length: int, lead_time: int = 0) -> None:
        assert (
            prediction_length > 0
        ), "The value of `prediction_length` should be > 0"
        assert lead_time >= 0, "The value of `lead_time` should be >= 0"

        self.prediction_length = prediction_length
        self.lead_time = lead_time

    def predict(self, dataset: Dataset, **kwargs) -> Iterator[Forecast]:
        """
        Compute forecasts for the time series in the provided dataset.
        This method is not implemented in this abstract class; please
        use one of the subclasses.
        Parameters
        ----------
        dataset
            The dataset containing the time series to predict.
        Returns
        -------
        Iterator[Forecast]
            Iterator over the forecasts, in the same order as the dataset
            iterable was provided.
        """
        raise NotImplementedError

    def serialize(self, path: Path) -> None:
        # serialize Predictor type
        with (path / "type.txt").open("w") as fp:
            fp.write(fqname_for(self.__class__))
        with (path / "version.json").open("w") as fp:
            json.dump(
                {"model": self.__version__, "gluonts": gluonts.__version__}, fp
            )

    @classmethod
    def deserialize(cls, path: Path, **kwargs) -> "Predictor":
        """
        Load a serialized predictor from the given path.

        Parameters
        ----------
        path
            Path to the serialized files predictor.
        **kwargs
            Optional context/device parameter to be used with the predictor.
            If nothing is passed will use the GPU if available and CPU
            otherwise.
        """
        # deserialize Predictor type
        with (path / "type.txt").open("r") as fp:
            tpe_str = fp.readline()

        tpe = locate(tpe_str)
        assert tpe is not None, f"Cannot locate {tpe_str}."

        # ensure that predictor_cls is a subtype of Predictor
        if not issubclass(tpe, Predictor):
            raise OSError(
                f"Class {fqname_for(tpe)} is not "
                f"a subclass of {fqname_for(Predictor)}"
            )

        # call deserialize() for the concrete Predictor type
        return tpe.deserialize(path, **kwargs)

    @classmethod
    def from_hyperparameters(cls, **hyperparameters):
        return from_hyperparameters(cls, **hyperparameters)

    @classmethod
    def derive_auto_fields(cls, train_iter):
        return {}

    @classmethod
    def from_inputs(cls, train_iter, **params):
        # auto_params usually include `use_feat_dynamic_real`,
        # `use_feat_static_cat` and `cardinality`
        auto_params = cls.derive_auto_fields(train_iter)
        # user specified 'params' will take precedence:
        params = {**auto_params, **params}
        return cls.from_hyperparameters(**params)


class RepresentablePredictor(Predictor):
    """
    An abstract predictor that can be subclassed by framework-specific models.
    Subclasses should have ``@validated()`` constructors:
    (de)serialization and equality test are all implemented on top of its logic.

    Parameters
    ----------
    prediction_length
        Prediction horizon.
    lead_time
        Prediction lead time.
    """

    def predict(self, dataset: Dataset, **kwargs) -> Iterator[Forecast]:
        for item in dataset:
            yield self.predict_item(item)

    def predict_item(self, item: DataEntry) -> Forecast:
        raise NotImplementedError

    def __eq__(self, that):
        """
        Two RepresentablePredictor instances are considered equal if they have
        the same constructor arguments.
        """
        return equals(self, that)

    def serialize(self, path: Path) -> None:
        # call Predictor.serialize() in order to serialize the class name
        super().serialize(path)
        with (path / "predictor.json").open("w") as fp:
            print(dump_json(self), file=fp)

    @classmethod
    def deserialize(cls, path: Path) -> "RepresentablePredictor":
        with (path / "predictor.json").open("r") as fp:
            return load_json(fp.read())


class WorkerError:
    def __init__(self, msg):
        self.msg = msg


def _worker_loop(
    predictor_path: Path,
    input_queue: mp.Queue,
    output_queue: mp.Queue,
    worker_id,
    **kwargs,
):
    """
    Worker loop for multiprocessing Predictor.

    Loads the predictor serialized in predictor_path reads inputs from
    input_queue and writes forecasts to output_queue
    """

    predictor = Predictor.deserialize(predictor_path)
    while True:
        idx, data_chunk = input_queue.get()
        if idx is None:
            output_queue.put((None, None, None))
            break
        try:
            result = list(predictor.predict(data_chunk, **kwargs))
        except Exception:
            we = WorkerError(
                "".join(traceback.format_exception(*sys.exc_info()))
            )
            output_queue.put((we, None, None))
            break
        output_queue.put((idx, worker_id, result))


class ParallelizedPredictor(Predictor):
    """
    Runs multiple instances (workers) of a predictor in parallel.
    Exceptions are propagated from the workers.
    Note: That there is currently an issue with tqdm that will cause things
    to hang if the ParallelizedPredictor is used with tqdm and an exception
    occurs during prediction.
    https://github.com/tqdm/tqdm/issues/548

    Parameters
    ----------
    base_predictor
        A representable predictor that will be used
    num_workers
        Number of workers (processes) to use. If set to
        None, one worker per CPU will be used.
    chunk_size
        Number of items to pass per call
    """

    def __init__(
        self,
        base_predictor: Predictor,
        num_workers: Optional[int] = None,
        chunk_size=1,
    ) -> None:
        super().__init__(
            lead_time=base_predictor.lead_time,
            prediction_length=base_predictor.prediction_length,
        )

        self._base_predictor = base_predictor
        self._num_workers = (
            num_workers if num_workers is not None else mp.cpu_count()
        )
        self._chunk_size = chunk_size
        self._num_running_workers = 0
        self._input_queues = []
        self._output_queue = None

    def _grouper(self, iterable, n):
        iterator = iter(iterable)
        group = tuple(itertools.islice(iterator, n))
        while group:
            yield group
            group = tuple(itertools.islice(iterator, n))

    def terminate(self):
        for q in self._input_queues:
            q.put((None, None))
        for w in self._workers:
            w.terminate()
        for i, w in enumerate(self._workers):
            w.join()

    def predict(self, dataset: Dataset, **kwargs) -> Iterator[Forecast]:
        with TemporaryDirectory() as tempdir:
            predictor_path = Path(tempdir)
            self._base_predictor.serialize(predictor_path)

            # TODO: Consider using shared memory for the data transfer.

            self._input_queues = [mp.Queue() for _ in range(self._num_workers)]
            self._output_queue = mp.Queue()

            workers = []
            for worker_id, in_q in enumerate(self._input_queues):
                worker = mp.Process(
                    target=_worker_loop,
                    args=(predictor_path, in_q, self._output_queue, worker_id),
                    kwargs=kwargs,
                )

                worker.daemon = True
                worker.start()
                workers.append(worker)
                self._num_running_workers += 1

            self._workers = workers

            chunked_data = self._grouper(dataset, self._chunk_size)

            self._send_idx = 0
            self._next_idx = 0

            self._data_buffer = {}

            worker_ids = list(range(self._num_workers))

            def receive():
                idx, worker_id, result = self._output_queue.get()
                if isinstance(idx, WorkerError):
                    self._num_running_workers -= 1
                    self.terminate()
                    raise Exception(idx.msg)
                if idx is not None:
                    self._data_buffer[idx] = result
                return idx, worker_id, result

            def get_next_from_buffer():
                while self._next_idx in self._data_buffer:
                    result_batch = self._data_buffer.pop(self._next_idx)
                    self._next_idx += 1
                    yield from result_batch

            def send(worker_id, chunk):
                q = self._input_queues[worker_id]
                q.put((self._send_idx, chunk))
                self._send_idx += 1

            try:
                # prime the queues
                for wid in worker_ids:
                    chunk = next(chunked_data)
                    send(wid, chunk)

                while True:
                    idx, wid, result = receive()
                    yield from get_next_from_buffer()
                    chunk = next(chunked_data)
                    send(wid, chunk)
            except StopIteration:
                # signal workers end of data
                for q in self._input_queues:
                    q.put((None, None))

            # collect any outstanding results
            while self._num_running_workers > 0:
                idx, worker_id, result = receive()
                if idx is None:
                    self._num_running_workers -= 1
                    continue
                yield from get_next_from_buffer()
            assert len(self._data_buffer) == 0
            assert self._send_idx == self._next_idx


class Localizer(Predictor):
    """
    A Predictor that uses an estimator to train a local model per time series
    and immediately calls this to predict.

    Parameters
    ----------
    estimator
        The estimator object to train on each dataset entry at prediction time.
    """

    def __init__(self, estimator: "Estimator"):
        super().__init__(
            lead_time=estimator.lead_time,
            prediction_length=estimator.prediction_length,
        )
        self.estimator = estimator

    def predict(self, dataset: Dataset, **kwargs) -> Iterator[Forecast]:
        logger = logging.getLogger(__name__)
        for i, ts in enumerate(dataset, start=1):
            logger.info(f"training for time series {i} / {len(dataset)}")
            trained_pred = self.estimator.train([ts])
            logger.info(f"predicting for time series {i} / {len(dataset)}")
            yield from trained_pred.predict([ts], **kwargs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Any

import mxnet as mx

from gluonts.core import fqname_for
from gluonts.core.serde import Kind, encode


@encode.register(mx.Context)
def encode_mx_context(v: mx.Context) -> Any:
    """
    Specializes :func:`encode` for invocations where ``v`` is an instance of
    the :class:`~mxnet.Context` class.
    """
    return {
        "__kind__": Kind.Instance,
        "class": fqname_for(v.__class__),
        "args": encode([v.device_type, v.device_id]),
    }


@encode.register(mx.nd.NDArray)
def encode_mx_ndarray(v: mx.nd.NDArray) -> Any:
    return {
        "__kind__": Kind.Instance,
        "class": "mxnet.nd.array",
        "args": encode([v.asnumpy().tolist()]),
        "kwargs": {"dtype": encode(v.dtype)},
    }



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import pytest

from gluonts.ext.r_forecast.util import (
    interval_to_quantile_level,
    quantile_to_interval_level,
)


@pytest.mark.parametrize(
    "quantile_level", [0.5, 0.15, 0.99, 0.4, 0.85, 0.35, 0.3, 0.5, 0.95, 0.13]
)
def test_quantile_to_interval_level_and_back(quantile_level: float):
    interval_level, side = quantile_to_interval_level(quantile_level)
    output = interval_to_quantile_level(interval_level, side)
    assert output == quantile_level



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pandas.tseries.frequencies import to_offset

from gluonts.time_feature import norm_freq_str


def test_norm_freq_str():
    assert norm_freq_str(to_offset("Y").name) == "A"
    assert norm_freq_str(to_offset("YS").name) == "A"
    assert norm_freq_str(to_offset("A").name) == "A"
    assert norm_freq_str(to_offset("AS").name) == "A"

    assert norm_freq_str(to_offset("Q").name) == "Q"
    assert norm_freq_str(to_offset("QS").name) == "Q"

    assert norm_freq_str(to_offset("M").name) == "M"
    assert norm_freq_str(to_offset("MS").name) == "M"

    assert norm_freq_str(to_offset("S").name) == "S"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List

import numpy as np
import pandas as pd
import pytest

from gluonts.evaluation import Evaluator, backtest_metrics
from gluonts.ext.r_forecast import (
    R_IS_INSTALLED,
    RPY2_IS_INSTALLED,
    RForecastPredictor,
)
from gluonts.model.forecast import QuantileForecast

# conditionally skip these tests if `R` and `rpy2` are not installed
if not R_IS_INSTALLED or not RPY2_IS_INSTALLED:
    skip_message = "Skipping test because `R` and `rpy2` are not installed!"
    pytest.skip(skip_message, allow_module_level=True)

freq = "H"
period = 24

## two weeks of data
dataset = [
    {
        "start": pd.Period("1990-01-01 00", freq=freq),
        "target": np.array(
            [
                item
                for i in range(70)
                for item in np.sin(
                    2 * np.pi / period * np.arange(1, period + 1, 1)
                )
            ]
        )
        + np.random.normal(0, 0.5, period * 70)
        + np.array(
            [
                item
                for i in range(10)
                for item in [0 for i in range(5 * 24)]
                + [8 for i in range(4)]
                + [0 for i in range(20)]
                + [8 for i in range(4)]
                + [0 for i in range(20)]
            ]
        ),
    }
]


def no_quantile_crossing(
    forecast: QuantileForecast, quantile_levels: List[float]
):
    sorted_levels = sorted(quantile_levels)
    quantile = forecast.quantile(sorted_levels[0])
    if np.isnan(quantile).any():
        return False

    for level in sorted_levels[1:]:
        prev_quantile = quantile
        quantile = forecast.quantile(level)
        if (quantile < prev_quantile).any():
            return False

    return True


@pytest.mark.parametrize(
    "method",
    [
        "ets",
        "arima",
        "fourier.arima",
    ],
)
def test_model_forecasts(method: str):
    prediction_length = 24 * 7
    quantile_levels = [0.5, 0.9, 0.2, 0.85]

    predictor = RForecastPredictor(
        freq=freq,
        prediction_length=prediction_length,
        period=period,
        method_name=method,
        params={
            "quantiles": quantile_levels,
            "output_types": ["mean", "quantiles"],
        },
    )

    forecast = list(predictor.predict(dataset))[0]

    assert forecast.prediction_length == prediction_length

    for level in quantile_levels:
        assert forecast.quantile(level).shape == (prediction_length,)

    assert no_quantile_crossing(forecast, quantile_levels)


def test_compare_arimas():
    evaluator = Evaluator(quantiles=[0.1, 0.5, 0.9])

    arima = RForecastPredictor(
        freq=freq,
        prediction_length=24 * 7,
        period=period,
        params={
            "quantiles": [0.50, 0.10, 0.90],
            "output_types": ["mean", "quantiles"],
        },
        method_name="arima",
    )

    arima_eval_metrics, _ = backtest_metrics(
        test_dataset=dataset, predictor=arima, evaluator=evaluator
    )

    fourier_arima = RForecastPredictor(
        freq=freq,
        prediction_length=24 * 7,
        period=period,
        params={
            "quantiles": [0.50, 0.10, 0.90],
            "output_types": ["mean", "quantiles"],
        },
        method_name="fourier.arima",
    )
    fourier_arima_eval_metrics, _ = backtest_metrics(
        test_dataset=dataset, predictor=fourier_arima, evaluator=evaluator
    )

    assert fourier_arima_eval_metrics["MASE"] < arima_eval_metrics["MASE"]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import tempfile
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from flaky import flaky
from pydantic import PositiveInt

from gluonts.dataset.artificial import constant_dataset
from gluonts.dataset.common import Dataset
from gluonts.dataset.util import forecast_start
from gluonts.evaluation import Evaluator, backtest_metrics
from gluonts.ext.naive_2 import Naive2Predictor
from gluonts.model.predictor import Predictor
from gluonts.model.seasonal_naive import SeasonalNaivePredictor


def generate_random_dataset(
    num_ts: int, start_time: str, freq: str, min_length: int, max_length: int
) -> Dataset:
    start_timestamp = pd.Period(start_time, freq=freq)
    for _ in range(num_ts):
        ts_length = np.random.randint(low=min_length, high=max_length)
        target = np.random.uniform(size=(ts_length,))
        data = {"target": target, "start": start_timestamp}
        yield data


PREDICTION_LENGTH = PositiveInt(30)
SEASON_LENGTH = PositiveInt(210)
START_TIME = "2018-01-03 14:37:12"  # That's a Wednesday
MIN_LENGTH = 300
MAX_LENGTH = 400
NUM_TS = 10


@pytest.mark.parametrize(
    "predictor_cls", [SeasonalNaivePredictor, Naive2Predictor]
)
@pytest.mark.parametrize(
    "freq", ["1min", "15min", "30min", "1H", "2H", "12H", "7D", "1W", "1M"]
)
def test_predictor(predictor_cls, freq: str):
    predictor = predictor_cls(
        freq=freq,
        prediction_length=PREDICTION_LENGTH,
        season_length=SEASON_LENGTH,
    )
    dataset = list(
        generate_random_dataset(
            num_ts=NUM_TS,
            start_time=START_TIME,
            freq=freq,
            min_length=MIN_LENGTH,
            max_length=MAX_LENGTH,
        )
    )

    # get forecasts
    forecasts = list(predictor.predict(dataset))

    assert len(dataset) == NUM_TS
    assert len(forecasts) == NUM_TS

    # check forecasts are as expected
    for data, forecast in zip(dataset, forecasts):
        assert forecast.samples.shape == (1, PREDICTION_LENGTH)

        ref = data["target"][
            -SEASON_LENGTH : -SEASON_LENGTH + PREDICTION_LENGTH
        ]

        assert forecast.start_date == forecast_start(data)

        # specifically for the seasonal naive we can test the supposed result directly
        if predictor_cls == SeasonalNaivePredictor:
            assert np.allclose(forecast.samples[0], ref)


# CONSTANT DATASET TESTS:


dataset_info, constant_train_ds, constant_test_ds = constant_dataset()
CONSTANT_DATASET_FREQ = dataset_info.metadata.freq
CONSTANT_DATASET_PREDICTION_LENGTH = dataset_info.prediction_length


def seasonal_naive_predictor():
    return (
        SeasonalNaivePredictor,
        dict(prediction_length=CONSTANT_DATASET_PREDICTION_LENGTH),
    )


def naive_2_predictor():
    return (
        Naive2Predictor,
        dict(prediction_length=CONSTANT_DATASET_PREDICTION_LENGTH),
    )


@flaky(max_runs=3, min_passes=1)
@pytest.mark.parametrize(
    "predictor_cls, parameters, accuracy",
    [seasonal_naive_predictor() + (0.0,), naive_2_predictor() + (0.0,)],
)
def test_accuracy(predictor_cls, parameters, accuracy):
    predictor = predictor_cls(freq=CONSTANT_DATASET_FREQ, **parameters)
    agg_metrics, item_metrics = backtest_metrics(
        test_dataset=constant_test_ds,
        predictor=predictor,
        evaluator=Evaluator(),
    )

    assert agg_metrics["ND"] <= accuracy


# SERIALIZATION/DESERIALIZATION TESTS:


@pytest.mark.parametrize(
    "predictor_cls, parameters",
    [seasonal_naive_predictor(), naive_2_predictor()],
)
def test_seriali_predictors(predictor_cls, parameters):
    predictor = predictor_cls(freq=CONSTANT_DATASET_FREQ, **parameters)
    with tempfile.TemporaryDirectory() as temp_dir:
        predictor.serialize(Path(temp_dir))
        predictor_exp = Predictor.deserialize(Path(temp_dir))
        assert predictor == predictor_exp



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset

from gluonts.core import serde

import gluonts.zebras as zb

FREQS = ["S", "min", "T", "H", "D", "B", "W", "M", "QS", "Q", "AS", "A", "Y"]


@pytest.mark.parametrize("freq", FREQS)
def test_freq_equal_pandas(freq):
    zb_freq = zb.freq(freq)
    pd_freq = to_offset(freq)

    assert zb_freq.to_pandas() == pd_freq

    freq = "7" + freq
    zb_freq = zb.freq(freq)
    pd_freq = to_offset(freq)

    assert zb_freq.to_pandas() == pd_freq


@pytest.mark.parametrize("freq", FREQS)
def test_freq_serde(freq):
    freq = zb.freq(freq)

    assert freq == serde.decode(serde.encode(freq))


@pytest.mark.parametrize("freq", FREQS)
def test_period_equal_pandas(freq):
    if freq != "S" and freq.endswith("S"):
        pytest.skip()

    pd_period = pd.Period("2020", freq=freq)
    zb_period = zb.period("2020", freq)

    assert pd_period == zb_period.to_pandas()
    assert pd_period + 9 == (zb_period + 9).to_pandas()

    freq = "7" + freq
    pd_period = pd.Period("2020", freq=freq)
    zb_period = zb.period("2020", freq)

    assert pd_period == zb_period.to_pandas()
    assert pd_period + 9 == (zb_period + 9).to_pandas()


@pytest.mark.parametrize("freq", FREQS)
def test_period_serde(freq):
    p = zb.period("2020", freq)

    assert p == serde.decode(serde.encode(p))


@pytest.mark.parametrize("freq", FREQS)
def test_periods_equal_pandas(freq):
    if freq != "S" and freq.endswith("S"):
        pytest.skip()

    pr = pd.period_range(start="2020", freq=freq, periods=20)
    ps = zb.periods("2020", freq, 20)

    np.testing.assert_array_equal(pr, ps.to_pandas())

    freq = "7" + freq
    pr = pd.period_range(start="2020", freq=freq, periods=20)
    ps = zb.periods("2020", freq, 20)

    np.testing.assert_array_equal(pr, ps.to_pandas())


@pytest.mark.parametrize("freq", FREQS)
def test_periods_serde(freq):
    ps = zb.periods("2020", freq, 20)

    assert ps[0] == serde.decode(serde.encode(ps[0]))
    assert ps == serde.decode(serde.encode(ps))


@pytest.mark.parametrize("freq", FREQS)
def test_periods_feature(freq):
    if freq != "S" and freq.endswith("S"):
        pytest.skip()

    pr = pd.period_range(start="2020", freq=freq, periods=20)
    ps = zb.periods("2020", freq, 20)

    np.testing.assert_array_equal(pr.start_time.year, ps.year)
    np.testing.assert_array_equal(pr.start_time.month, ps.month)
    np.testing.assert_array_equal(pr.start_time.day, ps.day)

    np.testing.assert_array_equal(pr.start_time.hour, ps.hour)
    np.testing.assert_array_equal(pr.start_time.minute, ps.minute)
    np.testing.assert_array_equal(pr.start_time.second, ps.second)

    np.testing.assert_array_equal(pr.start_time.dayofweek, ps.dayofweek)
    np.testing.assert_array_equal(pr.start_time.dayofyear, ps.dayofyear)
    np.testing.assert_array_equal(pr.start_time.isocalendar().week, ps.week)


@pytest.mark.parametrize(
    "date, freq, result",
    [
        (
            "2023-03-19",
            "W",
            ["2023-03-13", "2023-03-20", "2023-03-27", "2023-04-03"],
        ),
        (
            "2023-03-19",
            "W-SUN",
            ["2023-03-19", "2023-03-26", "2023-04-02", "2023-04-09"],
        ),
        (
            "2023-03-17",
            "W-SUN",
            ["2023-03-12", "2023-03-19", "2023-03-26", "2023-04-02"],
        ),
        (
            "2023-03-19",
            "W-SAT",
            ["2023-03-18", "2023-03-25", "2023-04-01", "2023-04-08"],
        ),
        (
            "2023-03-19",
            "2W-SAT",
            ["2023-03-18", "2023-04-01", "2023-04-15", "2023-04-29"],
        ),
        (
            "2023-03-17",
            "2W-SAT",
            ["2023-03-11", "2023-03-25", "2023-04-08", "2023-04-22"],
        ),
    ],
)
def test_weekly_weekday_period(date, freq, result):
    np.testing.assert_array_equal(
        zb.period(date, freq).periods(4).data,
        np.array(result).astype(np.datetime64),
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import functools
from typing import List, Optional, Type

import mxnet as mx
import numpy as np

from gluonts.dataset.common import DataBatch


def pad_to_size(
    x: np.array, size: int, axis: int = 0, is_right_pad: bool = True
):
    """
    Pads `xs` with 0 on the right (default) on the specified axis, which is the
    first axis by default.
    """

    pad_length = size - x.shape[axis]
    if pad_length <= 0:
        return x

    pad_width = [(0, 0)] * x.ndim
    right_pad = (0, pad_length)
    pad_width[axis] = right_pad if is_right_pad else right_pad[::-1]
    return np.pad(x, mode="constant", pad_width=pad_width)


def _is_stackable(arrays: List, axis: int = 0) -> bool:
    """
    Check if elements are scalars, have too few dimensions, or their target
    axes have equal length; i.e. they are directly `stack` able.
    """
    if isinstance(arrays[0], np.ndarray):
        s = {arr.shape[axis] for arr in arrays}
        return len(s) <= 1 and arrays[0].shape[axis] != 0
    return True


def _pad_arrays(
    data: List[np.ndarray],
    axis: int = 0,
    is_right_pad: bool = True,
) -> List[np.ndarray]:
    assert isinstance(data[0], np.ndarray)

    # MxNet causes a segfault when persisting 0-length arrays. As such,
    # we add a dummy pad of length 1 to 0-length dims.
    max_len = max(1, functools.reduce(max, (x.shape[axis] for x in data)))
    return [
        pad_to_size(x, max_len, axis, is_right_pad=is_right_pad) for x in data
    ]


def stack(
    data,
    ctx: Optional[mx.context.Context] = None,
    dtype: Optional[Type] = np.float32,
    variable_length: bool = False,
    is_right_pad: bool = True,
):
    if variable_length and not _is_stackable(data):
        data = _pad_arrays(data, axis=0, is_right_pad=is_right_pad)
    if isinstance(data[0], mx.nd.NDArray):
        # TODO: think about using shared context NDArrays
        #  https://github.com/awslabs/gluonts/blob/42bee73409f801e7bca73245ca21cd877891437c/src/gluonts/dataset/parallelized_loader.py#L157
        return mx.nd.stack(*data)
    if isinstance(data[0], np.ndarray):
        data = mx.nd.array(data, dtype=dtype, ctx=ctx)
    elif isinstance(data[0], (list, tuple)):
        return list(
            stack(
                t,
                ctx=ctx,
                dtype=dtype,
                variable_length=variable_length,
                is_right_pad=is_right_pad,
            )
            for t in zip(*data)
        )
    return data


def batchify(
    data: List[dict],
    ctx: Optional[mx.context.Context] = None,
    dtype: Optional[Type] = np.float32,
    variable_length: bool = False,
    is_right_pad: bool = True,
) -> DataBatch:
    return {
        key: stack(
            data=[item[key] for item in data],
            ctx=ctx,
            dtype=dtype,
            variable_length=variable_length,
            is_right_pad=is_right_pad,
        )
        for key in data[0].keys()
    }


def as_in_context(batch: dict, ctx: mx.Context = None) -> DataBatch:
    """
    Move data into new context, should only be in main process.
    """
    batch = {
        k: v.as_in_context(ctx) if isinstance(v, mx.nd.NDArray)
        # Workaround due to MXNet not being able to handle NDArrays with 0 in
        # shape properly:
        else (
            stack(v, ctx=ctx, dtype=v.dtype, variable_length=False)
            if isinstance(v[0], np.ndarray) and 0 in v[0].shape
            else v
        )
        for k, v in batch.items()
    }
    return batch



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Standard library imports
from typing import List, Tuple
import warnings

# Third-party imports
import mxnet as mx
from mxnet.ndarray import linalg_gemm2 as gemm2

# First-party imports
from gluonts.mx import Tensor
from gluonts.core.component import validated
from gluonts.mx.util import _broadcast_param
from gluonts.mx.activation import get_activation, get_activation_deriv
from gluonts.mx.block.sndense import SNDense


def jacobian_sn_mlp_block_bf(
    layers: List[Tuple[mx.gluon.HybridBlock, Tensor]]
) -> Tensor:
    """
    Brute force computation of the jacobian of a SNMlpBlock jac is of shape
    (Batch dim1, ..., Output dim, Input dim)

    Parameters
    ----------
    layers
        A list of tuples where each tuple (layer, input) is associated to a
        composing layer of the SNMLPBlock, where layer corresponds to the
        associated object layer, along with its input tensor.

    Returns
    -------
    Tensor
        Jacobian of the SNMLPBlock computed at a given input
    """
    for i, (layer, input) in enumerate(layers[::-1]):
        if isinstance(layer, SNDense):
            # broadcast weight of size (Output dim, Input dim)
            # to (Batch dim1, ..., Output dim, Input dim)
            jac_t = _broadcast_param(
                layer.weight,
                axes=range(len(input.shape[:-1])),
                sizes=input.shape[:-1],
            )
            if i == 0:
                jac = jac_t
            else:
                jac = gemm2(jac, jac_t)
        else:
            # act_deriv is of shape (Batch dim1, ..., Input dim)
            act_deriv = get_activation_deriv(layer)(mx.ndarray, input)
            # to (Batch dim1, ..., Output dim, Input dim) to fit the jacobian
            jac_t = act_deriv.expand_dims(len(jac.shape[:-2])).broadcast_axes(
                axis=len(jac.shape[:-2]), size=jac.shape[-2]
            )
            jac = jac * jac_t
    return jac


class SNMLPBlock(mx.gluon.HybridBlock):
    @validated()
    def __init__(
        self,
        in_units: int,
        hidden_units: int,
        out_units: int,
        num_hidden_layers: int = 2,
        activation: str = "lipswish",
        jacobian_method: str = "bf",
        num_power_iter: int = 1,
        coeff: float = 0.9,
        flatten: bool = False,
    ):
        super().__init__()
        self._in_units = in_units
        self._hidden_units = hidden_units
        self._out_units = out_units
        self._num_hidden_layers = num_hidden_layers
        self._activation = activation
        self._jacobian_method = jacobian_method
        self._num_power_iter = num_power_iter
        self._coeff = coeff
        self._weight_initializer = mx.init.Orthogonal(scale=self._coeff)
        self._bias_initializer = "zeros"
        self._flatten = flatten
        self._cached_inputs: List[Tensor] = []

        in_dim = self._in_units
        with self.name_scope():
            self._layers: List[mx.gluon.HybridBlock] = []
            for i in range(self._num_hidden_layers):
                lin = SNDense(
                    self._hidden_units,
                    in_units=in_dim,
                    activation=None,
                    num_power_iter=self._num_power_iter,
                    weight_initializer=self._weight_initializer,
                    bias_initializer=self._bias_initializer,
                    coeff=self._coeff,
                    flatten=self._flatten,
                )
                act = get_activation(
                    self._activation, prefix=self._activation + str(i)
                )
                in_dim = self._hidden_units
                self.register_child(lin)
                self.register_child(act)
                self._layers += [lin, act]

            last_lin = SNDense(
                self._out_units,
                in_units=in_dim,
                activation=None,
                num_power_iter=self._num_power_iter,
                weight_initializer=self._weight_initializer,
                bias_initializer=self._bias_initializer,
                coeff=self._coeff,
                flatten=self._flatten,
            )
            self.register_child(last_lin)
            self._layers += [last_lin]

    def get_weights(self):
        return [
            layer.weight
            for layer in self._layers
            if isinstance(layer, SNDense)
        ]

    def hybrid_forward(self, F, x: Tensor) -> Tensor:
        """

        Parameters
        ----------
        x
            Input Tensor

        Returns
        -------
        Tensor
            output of SNMLPBlock

        """
        self._cached_inputs = []
        for layer in self._layers:
            self._cached_inputs += [x]
            x = layer(x)
        return x

    def jacobian(self, x: Tensor) -> Tensor:
        """

        Parameters
        ----------
        x
            Input Tensor

        Returns
        -------
        Tensor
            Jacobian of the SNMLPBlock evaluated at x.

        """
        if self._jacobian_method == "ignore":
            return x * 0
        elif self._jacobian_method == "bf":
            if (
                len(self._cached_inputs) > 0
                and self._cached_inputs[0] is not x
            ):
                warnings.warn(
                    "Input not the same, recomputing forward for jacobian"
                    " term..."
                )
                self(x)
            return jacobian_sn_mlp_block_bf(
                [(l, i) for l, i in zip(self._layers, self._cached_inputs)]
            )
        raise NotImplementedError(self._jacobian_method)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
from typing import Callable, Optional

import numpy as np
import pandas as pd

from gluonts.itertools import batcher
from gluonts.core.component import validated
from gluonts.dataset.common import DataEntry, Dataset, ListDataset
from gluonts.dataset.field_names import FieldName


class MultivariateGrouper:
    """
    The MultivariateGrouper takes a univariate dataset and groups it into a
    single multivariate time series. Therefore, this class allows the user to
    convert a univariate dataset into a multivariate dataset without making a
    separate copy of the dataset.

    The Multivariate Grouper has two different modes:

    Training: For training data, the univariate time series get aligned to the
    earliest time stamp in the dataset. Time series will be left and right
    padded to produce an array of shape (dim, num_time_steps)

    Test: The test dataset might have multiple start dates (usually because
    the test dataset mimics a rolling evaluation scenario). In this case,
    the univariate dataset will be split into n multivariate time series,
    where n is the number of evaluation dates. Again, the
    time series will be grouped but only left padded. Note that the
    padded value will influence the prediction if the context length is
    longer than the length of the time series.

    Rules for padding for training and test datasets can be specified by the
    user.

    Parameters
    ----------
    max_target_dim
        Set maximum dimensionality (for faster testing or when hitting
        constraints of multivariate model). Takes the last max_target_dim
        time series and groups them to multivariate time series.
    num_test_dates
        Number of test dates in the test set. This can be more than one if
        the test set contains more than one forecast start date (often the
        case in a rolling evaluation scenario). Must be set to convert test
        data.
    train_fill_rule
        Implements the rule that fills missing data after alignment of the
        time series for the training dataset.
    test_fill_rule
        Implements the rule that fills missing data after alignment of the
        time series for the test dataset.
    """

    @validated()
    def __init__(
        self,
        max_target_dim: Optional[int] = None,
        num_test_dates: Optional[int] = None,
        train_fill_rule: Callable = np.mean,
        test_fill_rule: Callable = lambda x: 0.0,
    ) -> None:
        self.num_test_dates = num_test_dates
        self.max_target_dimension = max_target_dim
        self.train_fill_function = train_fill_rule
        self.test_fill_rule = test_fill_rule

        self.first_timestamp = None
        self.last_timestamp = None
        self.frequency = ""

    def __call__(self, dataset: Dataset) -> Dataset:
        self._preprocess(dataset)
        return self._group_all(dataset)

    def _preprocess(self, dataset: Dataset) -> None:
        """
        The preprocess function iterates over the dataset to gather data that
        is necessary for alignment.

        This includes     1) Storing first/last timestamp in the dataset     2)
        Storing the frequency of the dataset
        """
        for data in dataset:
            timestamp = data[FieldName.START]

            if self.first_timestamp is None:
                self.first_timestamp = timestamp

            if self.last_timestamp is None:
                self.last_timestamp = timestamp

            assert self.first_timestamp is not None
            assert self.last_timestamp is not None

            self.first_timestamp = min(self.first_timestamp, timestamp)
            self.last_timestamp = max(
                self.last_timestamp,
                timestamp + len(data[FieldName.TARGET]) - 1,
            )
            self.frequency = timestamp.freq

        logging.info(
            "first/last timestamp found: "
            f"{self.first_timestamp}/{self.last_timestamp}"
        )

    def _group_all(self, dataset: Dataset) -> Dataset:
        if self.num_test_dates is None:
            grouped_dataset = self._prepare_train_data(dataset)
        else:
            grouped_dataset = self._prepare_test_data(dataset)
        return grouped_dataset

    def _prepare_train_data(self, dataset: Dataset) -> Dataset:
        logging.info("group training time series to datasets")

        # Creates a single multivariate time series from the
        # univariate series in the dataset
        grouped_data = self._transform_target(self._align_data_entry, dataset)
        grouped_data[FieldName.TARGET] = np.vstack(
            grouped_data[FieldName.TARGET]
        )

        fields = next(iter(dataset), {}).keys()
        if FieldName.FEAT_DYNAMIC_REAL in fields:
            grouped_data[FieldName.FEAT_DYNAMIC_REAL] = np.vstack(
                [data[FieldName.FEAT_DYNAMIC_REAL] for data in dataset],
            )
        grouped_data = self._restrict_max_dimensionality(grouped_data)
        grouped_data[FieldName.START] = self.first_timestamp
        grouped_data[FieldName.FEAT_STATIC_CAT] = [0]

        return ListDataset(
            [grouped_data], freq=self.frequency, one_dim_target=False
        )

    def _prepare_test_data(self, dataset: Dataset) -> Dataset:
        assert self.num_test_dates is not None

        logging.info("group test time series to datasets")

        grouped_data = self._transform_target(self._left_pad_data, dataset)

        # Splits test dataset with rolling date into N R^d time series,
        # where N is the number of rolling evaluation dates
        assert len(grouped_data[FieldName.TARGET]) % self.num_test_dates == 0
        split_size = len(grouped_data[FieldName.TARGET]) // self.num_test_dates
        split_dataset = batcher(grouped_data[FieldName.TARGET], split_size)

        fields = next(iter(dataset), {}).keys()
        all_entries = list()
        for dataset_at_test_date in split_dataset:
            grouped_data = dict()
            grouped_data[FieldName.TARGET] = np.vstack(dataset_at_test_date)

            if FieldName.FEAT_DYNAMIC_REAL in fields:
                grouped_data[FieldName.FEAT_DYNAMIC_REAL] = np.vstack(
                    [data[FieldName.FEAT_DYNAMIC_REAL] for data in dataset],
                )
            grouped_data = self._restrict_max_dimensionality(grouped_data)
            grouped_data[FieldName.START] = self.first_timestamp
            grouped_data[FieldName.FEAT_STATIC_CAT] = [0]
            all_entries.append(grouped_data)

        return ListDataset(
            all_entries, freq=self.frequency, one_dim_target=False
        )

    def _align_data_entry(self, data: DataEntry) -> np.ndarray:
        ts = self.to_ts(data)
        return ts.reindex(
            pd.period_range(
                start=self.first_timestamp,
                end=self.last_timestamp,
                freq=data[FieldName.START].freq,
            ),
            fill_value=self.train_fill_function(ts),
        ).values

    def _left_pad_data(self, data: DataEntry) -> np.ndarray:
        ts = self.to_ts(data)
        return ts.reindex(
            pd.period_range(
                start=self.first_timestamp,
                end=ts.index[-1],
                freq=data[FieldName.START].freq,
            ),
            fill_value=self.test_fill_rule(ts),
        ).values

    @staticmethod
    def _transform_target(funcs, dataset: Dataset) -> DataEntry:
        return {FieldName.TARGET: [funcs(data) for data in dataset]}

    def _restrict_max_dimensionality(self, data: DataEntry) -> DataEntry:
        """
        Takes the last max_target_dimension dimensions from a multivariate
        dataentry.

        Parameters
        ----------
        data
            multivariate data entry with (dim, num_timesteps) target field

        Returns
        -------
        DataEntry
            data multivariate data entry with
            (max_target_dimension, num_timesteps) target field
        """

        if self.max_target_dimension is not None:
            # restrict maximum dimensionality (for faster testing)
            data[FieldName.TARGET] = data[FieldName.TARGET][
                -self.max_target_dimension :, :
            ]
            if FieldName.FEAT_DYNAMIC_REAL in data.keys():
                data[FieldName.FEAT_DYNAMIC_REAL] = data[
                    FieldName.FEAT_DYNAMIC_REAL
                ][-self.max_target_dimension :, :]
        return data

    @staticmethod
    def to_ts(data: DataEntry) -> pd.Series:
        return pd.Series(
            data[FieldName.TARGET],
            index=pd.period_range(
                start=data[FieldName.START],
                periods=len(data[FieldName.TARGET]),
                freq=data[FieldName.START].freq,
            ),
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
import sys

import pytest

from gluonts.mx.model.renewal import DeepRenewalProcessEstimator


@pytest.fixture()
def hyperparameters():
    return dict(
        num_cells=10,
        num_layers=3,
        context_length=10,
        epochs=1,
    )


@pytest.mark.skipif(sys.platform == "win32", reason="don't run on windows")
@pytest.mark.parametrize("hybridize", [True, False])
def test_accuracy_smoke_test(accuracy_test, hyperparameters, hybridize):
    hyperparameters.update(num_batches_per_epoch=80, hybridize=hybridize)
    accuracy_test(DeepRenewalProcessEstimator, hyperparameters, accuracy=10)


@pytest.mark.parametrize("hybridize", [True, False])
def test_serialize(serialize_test, hyperparameters, hybridize):
    hyperparameters.update(hybridize=hybridize)
    serialize_test(DeepRenewalProcessEstimator, hyperparameters)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, Optional

import numpy as np
import pandas as pd

from gluonts.core.component import validated
from gluonts.model.forecast import QuantileForecast

from . import RBasePredictor
from .util import (
    unlist,
    quantile_to_interval_level,
    interval_to_quantile_level,
)


UNIVARIATE_QUANTILE_FORECAST_METHODS = [
    "arima",
    "ets",
    "tbats",
    "thetaf",
    "stlar",
    "fourier.arima",
]
UNIVARIATE_POINT_FORECAST_METHODS = ["croston", "mlp"]
SUPPORTED_UNIVARIATE_METHODS = (
    UNIVARIATE_QUANTILE_FORECAST_METHODS + UNIVARIATE_POINT_FORECAST_METHODS
)


class RForecastPredictor(RBasePredictor):
    """
    Wrapper for calling the `R forecast package
    <http://pkg.robjhyndman.com/forecast/>`_.

    In order to use it you need to install R and rpy2. You also need the R `forecast` package which
    can be installed by running:

        R -e 'install.packages(c("forecast", "nnfor"), repos="https://cloud.r-project.org")' # noqa

    Parameters
    ----------
    freq
        The granularity of the time series (e.g. '1H')
    prediction_length
        Number of time points to be predicted.
    method_name
        The method from rforecast to be used one of
        "ets", "arima", "tbats", "croston", "mlp", "thetaf".
    period
        The period to be used (this is called `frequency` in the R forecast
        package), result to a tentative reasonable default if not specified
        (for instance 24 for hourly freq '1H')
    trunc_length
        Maximum history length to feed to the model (some models become slow
        with very long series).
    params
        Parameters to be used when calling the forecast method default.
        For `output_type`, 'mean' and `quantiles` are supported (depending
        on the underlying R method).
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        method_name: str = "ets",
        period: int = None,
        trunc_length: Optional[int] = None,
        save_info: bool = False,
        params: Dict = dict(),
    ) -> None:
        super().__init__(
            freq=freq,
            prediction_length=prediction_length,
            period=period,
            trunc_length=trunc_length,
            save_info=save_info,
            r_file_prefix="univariate",
        )
        assert method_name in SUPPORTED_UNIVARIATE_METHODS, (
            f"method {method_name} is not supported please "
            f"use one of {SUPPORTED_UNIVARIATE_METHODS}"
        )

        self.method_name = method_name
        self._r_method = self._robjects.r[method_name]

        self.params = {
            "prediction_length": self.prediction_length,
            "frequency": self.period,
        }

        if self.method_name in UNIVARIATE_POINT_FORECAST_METHODS:
            self.params["output_types"] = ["mean"]
        elif self.method_name in UNIVARIATE_QUANTILE_FORECAST_METHODS:
            self.params["output_types"] = ["mean", "quantiles"]
            self.params["intervals"] = list(range(0, 100, 10))

        if "quantiles" in params:
            assert (
                "intervals" not in params
            ), "Cannot specify both 'quantiles' and 'intervals'."
            intervals_info = [
                quantile_to_interval_level(ql) for ql in params["quantiles"]
            ]
            params["intervals"] = sorted(
                set([level for level, _ in intervals_info])
            )
            params.pop("quantiles")

        self.params.update(params)

        # Always ask for the mean prediction to be given,
        # since QuantileForecast will otherwise impute it
        # using the median, which is undesired.
        if "mean" not in self.params["output_types"]:
            self.params["output_types"].append("mean")

    def _get_r_forecast(self, data: Dict) -> Dict:
        make_ts = self._stats_pkg.ts
        r_params = self._robjects.vectors.ListVector(self.params)
        vec = self._robjects.FloatVector(data["target"])
        ts = make_ts(vec, frequency=self.period)
        forecast = self._r_method(ts, r_params)

        forecast_dict = dict(zip(forecast.names, map(unlist, list(forecast))))

        if "quantiles" in self.params["output_types"]:
            upper_quantiles = [
                str(interval_to_quantile_level(interval, side="upper"))
                for interval in self.params["intervals"]
            ]

            lower_quantiles = [
                str(interval_to_quantile_level(interval, side="lower"))
                for interval in self.params["intervals"]
            ]

            forecast_dict["quantiles"] = dict(
                zip(
                    lower_quantiles + upper_quantiles,
                    forecast_dict["lower_quantiles"]
                    + forecast_dict["upper_quantiles"],
                )
            )

        return forecast_dict

    def _preprocess_data(self, data: Dict) -> Dict:
        if self.trunc_length:
            shift_by = max(data["target"].shape[0] - self.trunc_length, 0)
            data["start"] = data["start"] + shift_by
            data["target"] = data["target"][-self.trunc_length :]
        return data

    def _warning_message(self) -> None:
        if self.method_name in UNIVARIATE_POINT_FORECAST_METHODS:
            print(
                "Overriding `output_types` to `mean` since"
                f" {self.method_name} is a point forecast method."
            )
        elif self.method_name in UNIVARIATE_QUANTILE_FORECAST_METHODS:
            print(
                "Overriding `output_types` to `quantiles` since "
                f"{self.method_name} is a quantile forecast method."
            )

    def _forecast_dict_to_obj(
        self,
        forecast_dict: Dict,
        forecast_start_date: pd.Timestamp,
        item_id: Optional[str],
        info: Dict,
    ) -> QuantileForecast:
        stats_dict = {"mean": forecast_dict["mean"]}

        if "quantiles" in forecast_dict:
            stats_dict.update(forecast_dict["quantiles"])

        forecast_arrays = np.array(list(stats_dict.values()))

        assert forecast_arrays.shape[1] == self.prediction_length

        return QuantileForecast(
            forecast_arrays=forecast_arrays,
            forecast_keys=list(stats_dict.keys()),
            start_date=forecast_start_date,
            item_id=item_id,
            info=info,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from .ts_stats import seasonal_error
from .stats import (
    absolute_label,
    error,
    absolute_error,
    squared_error,
    quantile_loss,
    coverage,
    absolute_percentage_error,
    symmetric_absolute_percentage_error,
    scaled_interval_score,
    absolute_scaled_error,
)
from .metrics import (
    Metric,
    mean_absolute_label,
    sum_absolute_label,
    SumAbsoluteError,
    MSE,
    SumQuantileLoss,
    Coverage,
    MAPE,
    SMAPE,
    MSIS,
    MASE,
    ND,
    RMSE,
    NRMSE,
    WeightedSumQuantileLoss,
    MAECoverage,
    MeanSumQuantileLoss,
    MeanWeightedSumQuantileLoss,
    OWA,
)
from .aggregations import Aggregation, Sum, Mean
from .evaluator import Evaluator, DirectEvaluator, DerivedEvaluator

__all__ = [
    "seasonal_error",
    "absolute_label",
    "error",
    "absolute_error",
    "squared_error",
    "quantile_loss",
    "coverage",
    "absolute_percentage_error",
    "symmetric_absolute_percentage_error",
    "scaled_interval_score",
    "absolute_scaled_error",
    "mean_absolute_label",
    "sum_absolute_label",
    "SumAbsoluteError",
    "MSE",
    "SumQuantileLoss",
    "Coverage",
    "MAPE",
    "SMAPE",
    "MSIS",
    "MASE",
    "ND",
    "RMSE",
    "NRMSE",
    "WeightedSumQuantileLoss",
    "MAECoverage",
    "MeanSumQuantileLoss",
    "MeanWeightedSumQuantileLoss",
    "OWA",
    "Aggregation",
    "Sum",
    "Mean",
    "Metric",
    "Evaluator",
    "DirectEvaluator",
    "DerivedEvaluator",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import re
import logging
from dataclasses import field
from typing import Callable, Dict, List, Optional, Union, Tuple

import numpy as np
import pandas as pd
from pydantic.dataclasses import dataclass

from gluonts.core.component import validated
from gluonts import maybe

logger = logging.getLogger(__name__)


def _linear_interpolation(
    xs: np.ndarray, ys: np.ndarray, x: float
) -> np.ndarray:
    assert sorted(xs) == xs
    assert len(xs) == len(ys)
    assert len(xs) >= 2

    if x < xs[0]:
        idx = 1
    elif x > xs[-1]:
        idx = len(xs) - 1
    else:
        idx = next(i for i, x_ in enumerate(xs) if x <= x_)

    x_low, x_high = xs[idx - 1 : idx + 1]
    y_low, y_high = ys[idx - 1 : idx + 1]

    dx = x_high - x_low
    dy = y_high - y_low

    return y_low + (x - x_low) / dx * dy


class ExponentialTailApproximation:
    """
    Approximate function on tails based on knots and make a inference on query
    point. Can be used for either interpolation or extrapolation on tails.

    Parameters
    ----------
    x_coord
        x-coordinates of the data points must be in increasing order.
    y_coord
        y-coordinates of the data points - may be a higher numpy array.
    tol
        tolerance when performing the division and computing the log in the
        exponential extrapolation.
    """

    def __init__(
        self,
        x_coord: List[float],
        y_coord: List[np.ndarray],
        tol: float = 1e-8,
    ) -> None:
        self.x_coord = x_coord
        assert sorted(self.x_coord) == self.x_coord
        self.y_coord = y_coord
        self.num_points = len(self.x_coord)
        assert (
            self.num_points >= 2
        ), "Need at least two points for exponential approximation."
        self.tol = tol
        (
            self.beta_inv_left,
            self.beta_inv_right,
        ) = self.init_exponential_tail_weights()

    def init_exponential_tail_weights(self) -> Tuple[float, float]:
        """
        Initialize the weight of exponentially decaying tail functions based on
        two extreme points on the left and right, respectively.

        Returns
        -------
        Tuple
            beta coefficient for left and right tails.
        """
        q_log_diff = np.log(
            (self.x_coord[1] + self.tol) / (self.x_coord[0] + self.tol)
            + self.tol
        )
        y_diff_left = self.y_coord[1] - self.y_coord[0]
        beta_inv_left = y_diff_left / q_log_diff

        z_log_diff = np.log(
            (1 - self.x_coord[-2] + self.tol)
            / (1 - self.x_coord[-1] + self.tol)
            + self.tol
        )  # z = 1/(1-q)
        y_diff_right = self.y_coord[-1] - self.y_coord[-2]
        beta_inv_right = y_diff_right / z_log_diff

        return beta_inv_left, beta_inv_right

    def left(self, x: float) -> np.ndarray:
        """
        Return the inference made on exponentially decaying tail functions.

        For left tail, x = exp(beta * (q - alpha))
        For right tail, x = 1 - exp(-beta * (q - alpha))

        E.g. for x = self.x_coord[0] or self.x_coord[1], return value is
        exactly self.y_coord[0] or self.y_coord[1], respectively.

        Parameters
        ----------
        x
            x-coordinate to evaluate the right tail.
        """
        return (
            self.beta_inv_left
            * np.log((x + self.tol) / (self.x_coord[1] + self.tol) + self.tol)
        ) + self.y_coord[1]

    def right(self, x: float) -> np.ndarray:
        """
        Return the inference made on exponentially decaying tail functions.

        For left tail, x = exp(beta * (q - alpha))
        For right tail, x = 1 - exp(-beta * (q - alpha))

        E.g. for x = self.x_coord[-1] or self.x_coord[-2] ,
        return value is exactly self.y_coord[-1]
        or self.y_coord[-2] respectively.
        Parameters
        ----------
        x
            x-coordinate to evaluate the right tail.
        """
        return (
            self.beta_inv_right
            * np.log(
                (1 - self.x_coord[-2] + self.tol) / (1 - x + self.tol)
                + self.tol
            )
        ) + self.y_coord[-2]

    def tail_range(self, default_left_tail=0.1, default_right_tail=0.9):
        """
        Return an effective range of left and right tails.
        """
        left_tail = max(
            self.x_coord[0],
            min(self.x_coord[1], default_left_tail),
        )
        right_tail = min(
            self.x_coord[-1],
            max(self.x_coord[-2], default_right_tail),
        )
        return left_tail, right_tail


@dataclass
class Quantile:
    value: float = field(metadata={"ge": 0.0, "le": 1.0})
    name: str

    @classmethod
    def from_float(cls, quantile: float) -> "Quantile":
        assert isinstance(quantile, float)
        return cls(value=quantile, name=str(quantile))

    @classmethod
    def from_str(cls, quantile: str) -> "Quantile":
        assert isinstance(quantile, str)

        try:
            return cls(value=float(quantile), name=quantile)
        except ValueError:
            m = re.match(r"^p(\d+)$", quantile)

            if m is None:
                raise ValueError(
                    'Quantile string should be of the form "p10", "p50", ...'
                    f' or "0.1", "0.5", ... but found {quantile}'
                )

            return cls.from_float(float(m.group(1)) / 100)

    @classmethod
    def parse(cls, quantile: Union["Quantile", float, str]) -> "Quantile":
        """
        Produces equivalent float and string representation of a given quantile
        level.

        >>> Quantile.parse(0.1)
        Quantile(value=0.1, name='0.1')

        >>> Quantile.parse('0.2')
        Quantile(value=0.2, name='0.2')

        >>> Quantile.parse('0.20')
        Quantile(value=0.2, name='0.20')

        >>> Quantile.parse('p99')
        Quantile(value=0.99, name='0.99')

        Parameters
        ----------
        quantile
            Quantile, can be a float a str representing a float e.g. '0.1' or a
            quantile string of the form 'p10'.

        Returns
        -------
        Quantile
            A tuple containing both a float and a string representation of the
            input quantile level.
        """
        if isinstance(quantile, Quantile):
            return quantile

        if isinstance(quantile, float):
            return cls.from_float(quantile)

        return cls.from_str(quantile)

    def __str__(self):
        return self.name


class Forecast:
    """
    A abstract class representing predictions.
    """

    start_date: pd.Period
    item_id: Optional[str]
    info: Optional[Dict]
    prediction_length: int
    mean: np.ndarray
    _index = None

    def quantile(self, q: Union[float, str]) -> np.ndarray:
        """
        Computes a quantile from the predicted distribution.

        Parameters
        ----------
        q
            Quantile to compute.

        Returns
        -------
        numpy.ndarray
            Value of the quantile across the prediction range.
        """
        raise NotImplementedError()

    def quantile_ts(self, q: Union[float, str]) -> pd.Series:
        return pd.Series(index=self.index, data=self.quantile(q))

    @property
    def median(self) -> np.ndarray:
        return self.quantile(0.5)

    @property
    def freq(self):
        return self.start_date.freq

    def __getitem__(self, name):
        if name == "mean":
            return self.mean
        elif name == "median":
            return self.median

        return self.quantile(name)

    def plot(
        self,
        *,
        intervals=(0.5, 0.9),
        ax=None,
        color=None,
        name=None,
        show_label=False,
    ):
        """
        Plot median forecast and prediction intervals using ``matplotlib``.

        By default the `0.5` and `0.9` prediction intervals are plotted. Other
        intervals can be choosen by setting `intervals`.

        This plots to the current axes object (via ``plt.gca()``), or to ``ax``
        if provided. Similarly, the color is using matplotlibs internal color
        cycle, if no explicit ``color`` is set.

        One can set ``name`` to use it as the ``label`` for the median
        forecast. Intervals are not labeled, unless ``show_label`` is set to
        ``True``.
        """
        import matplotlib.pyplot as plt

        # Get current axes (gca), if not provided explicitly.
        ax = maybe.unwrap_or_else(ax, plt.gca)

        # If no color is provided, we use matplotlib's internal color cycle.
        # Note: This is an internal API and might change in the future.
        color = maybe.unwrap_or_else(
            color, lambda: ax._get_lines.get_next_color()
        )

        # Plot median forecast
        ax.plot(
            self.index.to_timestamp(),
            self.quantile(0.5),
            color=color,
            label=name,
        )

        # Plot prediction intervals
        for interval in intervals:
            if show_label:
                if name is not None:
                    label = f"{name}: {interval}"
                else:
                    label = interval
            else:
                label = None

            # Translate interval to low and high values. E.g for `0.9` we get
            # `low = 0.05` and `high = 0.95`. (`interval + low + high == 1.0`)
            # Also, higher interval values mean lower confidence, and thus we
            # we use lower alpha values for them.
            low = (1 - interval) / 2
            ax.fill_between(
                # TODO: `index` currently uses `pandas.Period`, but we need
                # to pass a timestamp value to matplotlib. In the future this
                # will use ``zebras.Periods`` and thus needs to be adapted.
                self.index.to_timestamp(),
                self.quantile(low),
                self.quantile(1 - low),
                # Clamp alpha betwen ~16% and 50%.
                alpha=0.5 - interval / 3,
                facecolor=color,
                label=label,
            )

    @property
    def index(self) -> pd.PeriodIndex:
        if self._index is None:
            self._index = pd.period_range(
                self.start_date,
                periods=self.prediction_length,
                freq=self.start_date.freq,
            )
        return self._index

    def dim(self) -> int:
        """
        Returns the dimensionality of the forecast object.
        """
        raise NotImplementedError()

    def copy_dim(self, dim: int):
        """
        Returns a new Forecast object with only the selected sub-dimension.

        Parameters
        ----------
        dim
            The returned forecast object will only represent this dimension.
        """
        raise NotImplementedError()

    def copy_aggregate(self, agg_fun: Callable):
        """
        Returns a new Forecast object with a time series aggregated over the
        dimension axis.

        Parameters
        ----------
        agg_fun
            Aggregation function that defines the aggregation operation
            (typically mean or sum).
        """
        raise NotImplementedError()


class SampleForecast(Forecast):
    """
    A `Forecast` object, where the predicted distribution is represented
    internally as samples.

    Parameters
    ----------
    samples
        Array of size (num_samples, prediction_length) (1D case) or
        (num_samples, prediction_length, target_dim) (multivariate case)
    start_date
        start of the forecast
    info
        additional information that the forecaster may provide e.g. estimated
        parameters, number of iterations ran etc.
    """

    @validated()
    def __init__(
        self,
        samples: np.ndarray,
        start_date: pd.Period,
        item_id: Optional[str] = None,
        info: Optional[Dict] = None,
    ) -> None:
        assert isinstance(
            samples, np.ndarray
        ), "samples should be a numpy array"
        assert len(np.shape(samples)) == 2 or len(np.shape(samples)) == 3, (
            "samples should be a 2-dimensional or 3-dimensional array."
            " Dimensions found: {}".format(len(np.shape(samples)))
        )
        self.samples = samples
        self._sorted_samples_value = None
        self._mean = None
        self._dim = None
        self.item_id = item_id
        self.info = info

        assert isinstance(
            start_date, pd.Period
        ), "start_date should be a pandas Period object"
        self.start_date = start_date

    @property
    def _sorted_samples(self):
        if self._sorted_samples_value is None:
            self._sorted_samples_value = np.sort(self.samples, axis=0)
        return self._sorted_samples_value

    @property
    def num_samples(self):
        """
        The number of samples representing the forecast.
        """
        return self.samples.shape[0]

    @property
    def prediction_length(self):
        """
        Time length of the forecast.
        """
        return self.samples.shape[1]

    @property
    def mean(self) -> np.ndarray:
        """
        Forecast mean.
        """
        if self._mean is None:
            self._mean = np.mean(self.samples, axis=0)
        return self._mean

    @property
    def mean_ts(self) -> pd.Series:
        """
        Forecast mean, as a pandas.Series object.
        """
        return pd.Series(self.mean, index=self.index)

    def quantile(self, q: Union[float, str]) -> np.ndarray:
        q = Quantile.parse(q).value
        sample_idx = int(np.round((self.num_samples - 1) * q))
        return self._sorted_samples[sample_idx, :]

    def copy_dim(self, dim: int) -> "SampleForecast":
        if len(self.samples.shape) == 2:
            samples = self.samples
        else:
            target_dim = self.samples.shape[2]
            assert dim < target_dim, (
                f"must set 0 <= dim < target_dim, but got dim={dim},"
                f" target_dim={target_dim}"
            )
            samples = self.samples[:, :, dim]

        return SampleForecast(
            samples=samples,
            start_date=self.start_date,
            item_id=self.item_id,
            info=self.info,
        )

    def copy_aggregate(self, agg_fun: Callable) -> "SampleForecast":
        if len(self.samples.shape) == 2:
            samples = self.samples
        else:
            # Aggregate over target dimension axis
            samples = agg_fun(self.samples, axis=2)
        return SampleForecast(
            samples=samples,
            start_date=self.start_date,
            item_id=self.item_id,
            info=self.info,
        )

    def dim(self) -> int:
        if self._dim is None:
            if len(self.samples.shape) == 2:
                # univariate target
                # shape: (num_samples, prediction_length)
                self._dim = 1
            else:
                # multivariate target
                # shape: (num_samples, prediction_length, target_dim)
                self._dim = self.samples.shape[2]
        return self._dim

    def __repr__(self):
        return ", ".join(
            [
                f"SampleForecast({self.samples!r})",
                f"{self.start_date!r}",
                f"item_id={self.item_id!r}",
                f"info={self.info!r})",
            ]
        )

    def to_quantile_forecast(self, quantiles: List[str]) -> "QuantileForecast":
        return QuantileForecast(
            forecast_arrays=np.array(
                [
                    self.quantile(q) if q != "mean" else self.mean()
                    for q in quantiles
                ]
            ),
            start_date=self.start_date,
            forecast_keys=quantiles,
            item_id=self.item_id,
            info=self.info,
        )


class QuantileForecast(Forecast):
    """
    A Forecast that contains arrays (i.e. time series) for quantiles and mean.

    Parameters
    ----------
    forecast_arrays
        An array of forecasts
    start_date
        start of the forecast
    forecast_keys
        A list of quantiles of the form '0.1', '0.9', etc.,
        and potentially 'mean'. Each entry corresponds to one array in
        forecast_arrays.
    info
        additional information that the forecaster may provide e.g. estimated
        parameters, number of iterations ran etc.
    """

    def __init__(
        self,
        forecast_arrays: np.ndarray,
        start_date: pd.Period,
        forecast_keys: List[str],
        item_id: Optional[str] = None,
        info: Optional[Dict] = None,
    ) -> None:
        self.forecast_array = forecast_arrays
        assert isinstance(
            start_date, pd.Period
        ), "start_date should be a pandas Period object"
        self.start_date = start_date

        # normalize keys
        self.forecast_keys = [
            Quantile.from_str(key).name if key != "mean" else key
            for key in forecast_keys
        ]
        self.item_id = item_id
        self.info = info
        self._dim = None

        shape = self.forecast_array.shape
        assert shape[0] == len(self.forecast_keys), (
            f"The forecast_array (shape={shape} should have the same "
            f"length as the forecast_keys (len={len(self.forecast_keys)})."
        )
        self.prediction_length = shape[1]
        self._forecast_dict = {
            k: self.forecast_array[i] for i, k in enumerate(self.forecast_keys)
        }
        self._nan_out = np.array([np.nan] * self.prediction_length)

    def quantile(self, inference_quantile: Union[float, str]) -> np.ndarray:
        sorted_forecast_dict = dict(sorted(self._forecast_dict.items()))
        sorted_forecast_dict.pop("mean", None)
        quantiles = [float(q) for q in sorted_forecast_dict.keys()]
        quantile_predictions = list(sorted_forecast_dict.values())

        inference_quantile = Quantile.parse(inference_quantile).value

        if len(quantiles) < 2 or inference_quantile in quantiles:
            q_str = Quantile.parse(inference_quantile).name
            return self._forecast_dict.get(q_str, self._nan_out)

        exp_tail_approximation = ExponentialTailApproximation(
            quantiles, quantile_predictions
        )
        # The effective range of left, right tails varies over tail
        # approximation class
        (
            left_tail_quantile,
            right_tail_quantile,
        ) = exp_tail_approximation.tail_range()

        if inference_quantile <= left_tail_quantile:
            return exp_tail_approximation.left(inference_quantile)
        elif inference_quantile >= right_tail_quantile:
            return exp_tail_approximation.right(inference_quantile)
        else:
            return _linear_interpolation(
                quantiles, quantile_predictions, inference_quantile
            )

    def copy_dim(self, dim: int) -> "QuantileForecast":
        if len(self.forecast_array.shape) == 2:
            forecast_array = self.forecast_array
        else:
            target_dim = self.forecast_array.shape[2]
            assert dim < target_dim, (
                f"must set 0 <= dim < target_dim, but got dim={dim},"
                f" target_dim={target_dim}"
            )
            forecast_array = self.forecast_array[:, :, dim]

        return QuantileForecast(
            forecast_arrays=forecast_array,
            start_date=self.start_date,
            forecast_keys=self.forecast_keys,
            item_id=self.item_id,
            info=self.info,
        )

    @property
    def mean(self) -> np.ndarray:
        """
        Forecast mean.
        """
        if "mean" in self._forecast_dict:
            return self._forecast_dict["mean"]
        logger.warning(
            "The mean prediction is not stored in the forecast data; "
            "the median is being returned instead. "
            "This behaviour may change in the future."
        )
        return self.quantile("p50")

    def dim(self) -> int:
        if self._dim is None:
            if len(self.forecast_array.shape) == 2:
                # univariate target
                # shape: (num_samples, prediction_length)
                self._dim = 1
            else:
                # multivariate target
                # shape: (num_samples, prediction_length, target_dim)
                self._dim = self.forecast_array.shape[2]
        return self._dim

    def __repr__(self):
        return ", ".join(
            [
                f"QuantileForecast({self.forecast_array!r})",
                f"start_date={self.start_date!r}",
                f"forecast_keys={self.forecast_keys!r}",
                f"item_id={self.item_id!r}",
                f"info={self.info!r})",
            ]
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Tuple

import mxnet as mx
import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset
import pytest
from mxnet import nd

from gluonts.mx import PointProcessGluonPredictor, PointProcessSampleForecast

from gluonts.mx import Tensor
from gluonts.transform import (
    ContinuousTimeInstanceSplitter,
    ContinuousTimePredictionSampler,
)

from .common import point_process_dataset, point_process_dataset_2


class MockTPPPredictionNet(mx.gluon.HybridBlock):
    def __init__(
        self,
        num_parallel_samples: int = 100,
        prediction_interval_length: float = 5.0,
        context_interval_length: float = 5.0,
    ) -> None:
        super().__init__()
        self.num_parallel_samples = num_parallel_samples
        self.prediction_interval_length = prediction_interval_length
        self.context_interval_length = context_interval_length

    def hybridize(self, active=True, **kwargs):
        if active:
            raise NotImplementedError()

    def hybrid_forward(
        self, F, past_target: Tensor, past_valid_length: Tensor
    ) -> Tuple[Tensor, Tensor]:
        """
        Return two tensors, of shape
        (batch_size, num_samples, max_prediction_length, target_dim)
        and (batch_size, num_samples) respectively.
        """
        batch_size = past_target.shape[0]
        assert past_valid_length.shape[0] == batch_size

        target_shape = (self.num_parallel_samples, batch_size, 25)
        pred_target = nd.stack(
            nd.random.uniform(shape=target_shape),
            nd.random.randint(0, 10, shape=target_shape).astype(np.float32),
            axis=-1,
        )
        pred_valid_length = nd.random.randint(
            15, 25 + 1, shape=target_shape[:2]
        )

        return pred_target, pred_valid_length


@pytest.fixture
def predictor_factory():
    def get_predictor(**kwargs) -> PointProcessGluonPredictor:
        default_kwargs = dict(
            input_names=["past_target", "past_valid_length"],
            prediction_net=MockTPPPredictionNet(
                prediction_interval_length=5.0
            ),
            batch_size=128,
            prediction_interval_length=5.0,
            freq="H",
            ctx=mx.cpu(),
            input_transform=ContinuousTimeInstanceSplitter(
                1,
                5,
                to_offset("H"),
                ContinuousTimePredictionSampler(
                    allow_empty_interval=False, min_past=1
                ),
            ),
        )

        default_kwargs.update(**kwargs)

        return PointProcessGluonPredictor(**default_kwargs)

    return get_predictor


@pytest.mark.parametrize(
    "dataset_tuple", [(point_process_dataset, 1), (point_process_dataset_2, 3)]
)
def test_tpp_pred_dataset_2_shapes_ok(dataset_tuple, predictor_factory):
    dataset, ds_length = dataset_tuple

    predictor = predictor_factory()
    forecasts = [fc for fc in predictor.predict(dataset(), 50)]

    assert len(forecasts) == ds_length

    for forecast in forecasts:
        # each forecast should have 3dim samples, 1d valid lengths
        assert isinstance(forecast, PointProcessSampleForecast)

        assert forecast.samples.shape == (50, 25, 2)
        assert forecast.valid_length.shape == (50,)

        assert (
            forecast.prediction_interval_length
            == predictor.prediction_interval_length
        )

        assert forecast.start_date == pd.Timestamp("2011-01-01 03:00:00")
        assert forecast.end_date == pd.Timestamp("2011-01-01 08:00:00")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from mxnet.gluon import HybridBlock, rnn

from gluonts.core.component import validated
from gluonts.mx import Tensor


class RNN(HybridBlock):
    """
    Defines an RNN block.

    Parameters
    ----------
    mode
        type of the RNN. Can be either: rnn_relu (RNN with relu activation),
        rnn_tanh, (RNN with tanh activation), lstm or gru.

    num_hidden
        number of units per hidden layer.

    num_layers
        number of hidden layers.

    bidirectional
        toggle use of bi-directional RNN as encoder.
    """

    @validated()
    def __init__(
        self,
        mode: str,
        num_hidden: int,
        num_layers: int,
        bidirectional: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        with self.name_scope():
            if mode == "rnn_relu":
                self.rnn = rnn.RNN(
                    num_hidden,
                    num_layers,
                    bidirectional=bidirectional,
                    activation="relu",
                    layout="NTC",
                )
            elif mode == "rnn_tanh":
                self.rnn = rnn.RNN(
                    num_hidden,
                    num_layers,
                    bidirectional=bidirectional,
                    layout="NTC",
                )
            elif mode == "lstm":
                self.rnn = rnn.LSTM(
                    num_hidden,
                    num_layers,
                    bidirectional=bidirectional,
                    layout="NTC",
                )
            elif mode == "gru":
                self.rnn = rnn.GRU(
                    num_hidden,
                    num_layers,
                    bidirectional=bidirectional,
                    layout="NTC",
                )
            else:
                raise ValueError(
                    "Invalid mode %s. Options are rnn_relu, rnn_tanh, lstm,"
                    " and gru " % mode
                )

    def hybrid_forward(self, F, inputs: Tensor) -> Tensor:  # NTC in, NTC out
        """

        Parameters
        ----------
        F
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.

        inputs
            input tensor with shape (batch_size, num_timesteps, num_dimensions)

        Returns
        -------
        Tensor
            rnn output with shape (batch_size, num_timesteps, num_dimensions)
        """
        return self.rnn(inputs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from gluonts.shell.sagemaker.nested_params import decode_nested_parameters


def test_nested_params():
    data = decode_nested_parameters(
        {
            "$env.num_workers": "4",
            "$evaluation.quantiles": [0.1, 0.5, 0.9],
            "prediction_length": 14,
        }
    )

    hps = data.pop("")
    assert hps["prediction_length"] == 14

    env = data.pop("env")
    assert env["num_workers"] == "4"

    evaluation = data.pop("evaluation")
    assert evaluation["quantiles"] == [0.1, 0.5, 0.9]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional, Tuple

import mxnet as mx
from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.distribution import Distribution, DistributionOutput
from gluonts.mx.distribution.distribution import getF
from mxnet import gluon


class DeepRenewalNetwork(gluon.HybridBlock):
    def __init__(
        self,
        prediction_length: int,
        context_length: int,
        interval_distr_output: DistributionOutput,
        size_distr_output: DistributionOutput,
        num_cells: int,
        num_layers: int,
        dropout_rate: float = 0.1,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.prediction_length = prediction_length
        self.context_length = context_length
        self.interval_distr_output = interval_distr_output
        self.size_distr_output = size_distr_output
        self.num_cells = num_cells
        self.num_layers = num_layers

        with self.name_scope():
            self.rnn_cell = mx.gluon.rnn.LSTM(
                hidden_size=num_cells,
                num_layers=num_layers,
                layout="NTC",
                dropout=dropout_rate,
                h2h_weight_initializer=mx.init.Xavier(),
                h2r_weight_initializer=mx.init.Xavier(),
            )
            self.dense = mx.gluon.nn.Dense(
                units=2,
                activation="softrelu",
                flatten=False,
                weight_initializer=mx.init.Xavier(),
                bias_initializer=mx.init.Constant(5.0),
            )

        # add constant learnable parameters if the distributions require
        # more than the conditional mean
        with self.name_scope():
            if len(self.interval_distr_output.args_dim) > 1:
                self.interval_alpha_bias = self.params.get(
                    "interval_alpha_bias",
                    shape=(1,),
                    init=mx.init.Constant(2),
                )
            if len(self.size_distr_output.args_dim) > 1:
                self.size_alpha_bias = self.params.get(
                    "size_alpha_bias",
                    shape=(1,),
                    init=mx.init.Constant(2),
                )

    def begin_state(self, batch_size: int) -> Tuple[Tensor, Tensor]:
        return self.rnn_cell.begin_state(batch_size=batch_size)

    def distribution(
        self,
        cond_mean: Tensor,
        interval_alpha_bias: Optional[Tensor] = None,
        size_alpha_bias: Optional[Tensor] = None,
    ) -> Tuple[Distribution, ...]:
        F = getF(cond_mean)

        cond_interval, cond_size = F.split(cond_mean, num_outputs=2, axis=-1)

        alpha_biases = [
            F.broadcast_mul(F.ones_like(cond_interval), bias)
            if bias is not None
            else None
            for bias in [interval_alpha_bias, size_alpha_bias]
        ]

        distr_params = zip(
            [self.interval_distr_output, self.size_distr_output],
            [cond_interval, cond_size],
            alpha_biases,
        )

        return tuple(
            (
                do.distribution(mean)
                if len(do.args_dim) == 1
                else do.distribution(
                    [mean, F.Activation(alpha_bias, "softrelu") + 1e-5]
                )
            )
            for ix, (do, mean, alpha_bias) in enumerate(distr_params)
        )

    @staticmethod
    def forwardshift(A):
        """
        Shift an array's content forward by 1 time step along the first axis,
        keeping the shape identical by padding on the left with zeros.

        Parameters
        ----------
        A : nd.NDArray
            Shape (N, T, ...), the tensor in which the entries will be shifted
            forward by one
        """
        F = getF(A)
        A = F.Concat(
            F.zeros_like(F.slice_axis(A, axis=1, begin=0, end=1)), A, dim=1
        )
        return F.slice_axis(A, axis=1, begin=0, end=-1)

    def mu_map(
        self,
        data: Tensor,
        shift: bool = True,
        state=None,
    ) -> Tensor:
        """
        Map a given (N, T, 2) tensor to conditional interval and size means of
        the next time step.

        Parameters
        ----------
        data: Tensor
            tensor of shape (N, T, 2) containing the past target, with the
            first array along the last dimension containing intervals, and
            the second array containing demand sizes.
        shift: bool
            if True, the past data will be shifted forward by one time step
        state
            RNN cell state if available

        Returns
        -------
        conditional_means: Tensor
            tensor of shape (N, T, 2) containing conditional means for
            intervals and sizes respectively
        rnn_out_state
            output state of the RNN, returned only if the input `state` is not
            None
        """
        data = self.forwardshift(data) if shift else data

        if state is not None:
            rnn_out, rnn_out_state = self.rnn_cell(data, state)
        else:
            rnn_out = self.rnn_cell(data)

        dense_out = self.dense(rnn_out)
        model_mu = dense_out + 1e-5

        return model_mu if state is None else (model_mu, rnn_out_state)


class DeepRenewalTrainingNetwork(DeepRenewalNetwork):
    def hybrid_forward(
        self,
        F,
        past_target,
        valid_length,
        interval_alpha_bias=None,
        size_alpha_bias=None,
        **kwargs,
    ) -> Tensor:
        """
        Compute negative log likelihood losses for given data.

        Parameters
        ----------
        F
        past_target
            shape (batch_size, history_length, 2). Expects a right-aligned
            ragged tensor containing the past data in an interval-size
            format.
        valid_length
            number of valid data points in each array of the batch. i.e.,
            non-valid points will be masked out. shape (batch_size, 1)
        interval_alpha_bias
        size_alpha_bias

        Returns
        -------
        loss
            negative log likelihood with shape (batch_size, history_length)
        """
        cond_mean = self.mu_map(past_target, shift=True)
        dist_interval, dist_size = self.distribution(
            cond_mean, interval_alpha_bias, size_alpha_bias
        )
        data_interval, data_size = F.split(past_target, num_outputs=2, axis=-1)

        # TODO: on windows, operators below may produce NaN values

        log_prob = F.squeeze(
            dist_interval.log_prob(data_interval - 1)
            + dist_size.log_prob(data_size - 1),
            axis=-1,
        )
        reverse_lengths = F.broadcast_sub(
            F.maximum(1, F.max(valid_length)), valid_length
        ).squeeze(-1)
        mask = F.SequenceMask(
            F.ones_like(log_prob),
            reverse_lengths,
            use_sequence_length=True,
            axis=1,
        )
        loss = -F.where(1 - mask, log_prob, F.zeros_like(log_prob))

        return loss


class DeepRenewalPredictionNetwork(DeepRenewalNetwork):
    @validated()
    def __init__(self, num_parallel_samples: int = 100, **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_parallel_samples = num_parallel_samples

    def sampling_decoder(
        self,
        F,
        past_target: Tensor,
        interval_alpha_bias: Optional[Tensor] = None,
        size_alpha_bias: Optional[Tensor] = None,
        time_remaining: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Sample a trajectory of interval-size pairs from the model given past
        target.

        Parameters
        ----------
        past_target
            (batch_size, history_length, 2) shaped tensor containing the past
            time series in an interval-size representation.
        time_remaining
            (batch_size, 1) shaped tensor containing the number of time steps
            that were zero before the start of the forecast horizon

        Returns
        -------
        samples
            samples of shape (batch_size, num_parallel_samples, 2,
            sequence_length)
        """
        if time_remaining is None:
            time_remaining = F.broadcast_like(
                F.zeros((1,)), past_target, lhs_axes=(0,), rhs_axes=(0,)
            )

        # if no valid points in the past_target, override time remaining with
        # zero
        batch_sum = F.sum(past_target, axis=1).sum(-1).expand_dims(-1)
        time_remaining = F.where(
            batch_sum > 0, time_remaining, F.zeros_like(time_remaining)
        )

        repeated_past_target = past_target.repeat(
            repeats=self.num_parallel_samples, axis=0
        )  # (N * samples, T, 2)
        repeated_time_remaining = time_remaining.repeat(
            repeats=self.num_parallel_samples, axis=0
        )  # (N * samples, 1)

        interval_samples = []
        size_samples = []

        for t in range(self.prediction_length):
            cond_mean = self.mu_map(repeated_past_target, shift=False)
            cond_mean_last = F.slice_axis(
                cond_mean, axis=1, begin=-1, end=None
            ).squeeze(1)

            dist_interval, dist_size = self.distribution(
                cond_mean_last, interval_alpha_bias, size_alpha_bias
            )

            # initial samples for interval should be taken conditionally
            # we achieve this via (leaky) rejection sampling
            if t == 0:
                interval_sample = F.zeros_like(repeated_time_remaining)
                for j in range(50):
                    interval_sample = F.where(
                        interval_sample > repeated_time_remaining,
                        interval_sample,
                        dist_interval.sample(),
                    )
                interval_sample = (
                    F.where(
                        interval_sample <= repeated_time_remaining,
                        repeated_time_remaining,
                        interval_sample,
                    )
                    + 1
                )
            else:
                interval_sample = dist_interval.sample() + 1
            size_sample = dist_size.sample() + 1

            interval_samples.append(interval_sample)
            size_samples.append(size_sample)

            repeated_past_target = F.concat(
                repeated_past_target,
                F.concat(interval_sample, size_sample, dim=-1).expand_dims(1),
                dim=1,
            )

        interval_samples[0] = interval_samples[0] - repeated_time_remaining
        samples = F.concat(
            *[
                F.concat(x, y, dim=-1).expand_dims(1)
                for x, y in zip(interval_samples, size_samples)
            ],
            dim=1,
        )

        return samples.reshape(
            shape=(-1, self.num_parallel_samples) + samples.shape[1:]
        ).swapaxes(2, 3)

    def hybrid_forward(
        self,
        F,
        past_target,
        time_remaining,
        interval_alpha_bias=None,
        size_alpha_bias=None,
        **kwargs,
    ) -> Tensor:
        return self.sampling_decoder(
            F,
            past_target=past_target,
            interval_alpha_bias=interval_alpha_bias,
            size_alpha_bias=size_alpha_bias,
            time_remaining=time_remaining,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
import shutil
from functools import partial
from pathlib import Path
from typing import Any, Dict, Optional

from gluonts.dataset import DatasetWriter
from gluonts.dataset.artificial import ConstantDataset
from gluonts.dataset.common import TrainDatasets, load_datasets
from gluonts.dataset.jsonl import JsonLinesWriter

from ._artificial import generate_artificial_dataset
from ._airpassengers import generate_airpassengers_dataset
from ._gp_copula_2019 import generate_gp_copula_dataset
from ._lstnet import generate_lstnet_dataset
from ._m3 import generate_m3_dataset
from ._m4 import generate_m4_dataset
from ._m5 import generate_m5_dataset
from ._tsf_datasets import generate_forecasting_dataset
from ._uber_tlc import generate_uber_dataset


def get_download_path() -> Path:
    """

    Returns
    -------
    Path
        default path to download datasets or models of gluon-ts.
        The path is $HOME/.gluonts/
    """
    return Path.home() / ".gluonts"


dataset_recipes = {
    # each recipe generates a dataset given a path
    "constant": partial(
        generate_artificial_dataset, dataset=ConstantDataset()
    ),
    "exchange_rate": partial(
        generate_lstnet_dataset, dataset_name="exchange_rate"
    ),
    "solar-energy": partial(
        generate_lstnet_dataset, dataset_name="solar-energy"
    ),
    "electricity": partial(
        generate_lstnet_dataset, dataset_name="electricity"
    ),
    "traffic": partial(generate_lstnet_dataset, dataset_name="traffic"),
    "exchange_rate_nips": partial(
        generate_gp_copula_dataset, dataset_name="exchange_rate_nips"
    ),
    "electricity_nips": partial(
        generate_gp_copula_dataset, dataset_name="electricity_nips"
    ),
    "traffic_nips": partial(
        generate_gp_copula_dataset, dataset_name="traffic_nips"
    ),
    "solar_nips": partial(
        generate_gp_copula_dataset, dataset_name="solar_nips"
    ),
    "wiki2000_nips": partial(
        generate_gp_copula_dataset, dataset_name="wiki2000_nips"
    ),
    "wiki-rolling_nips": partial(
        generate_gp_copula_dataset, dataset_name="wiki-rolling_nips"
    ),
    "taxi_30min": partial(
        generate_gp_copula_dataset, dataset_name="taxi_30min"
    ),
    "kaggle_web_traffic_with_missing": partial(
        generate_forecasting_dataset,
        dataset_name="kaggle_web_traffic_with_missing",
    ),
    "kaggle_web_traffic_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="kaggle_web_traffic_without_missing",
    ),
    "kaggle_web_traffic_weekly": partial(
        generate_forecasting_dataset,
        dataset_name="kaggle_web_traffic_weekly",
    ),
    "m1_yearly": partial(
        generate_forecasting_dataset, dataset_name="m1_yearly"
    ),
    "m1_quarterly": partial(
        generate_forecasting_dataset, dataset_name="m1_quarterly"
    ),
    "m1_monthly": partial(
        generate_forecasting_dataset, dataset_name="m1_monthly"
    ),
    "nn5_daily_with_missing": partial(
        generate_forecasting_dataset, dataset_name="nn5_daily_with_missing"
    ),
    "nn5_daily_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="nn5_daily_without_missing",
    ),
    "nn5_weekly": partial(
        generate_forecasting_dataset, dataset_name="nn5_weekly"
    ),
    "tourism_monthly": partial(
        generate_forecasting_dataset, dataset_name="tourism_monthly"
    ),
    "tourism_quarterly": partial(
        generate_forecasting_dataset, dataset_name="tourism_quarterly"
    ),
    "tourism_yearly": partial(
        generate_forecasting_dataset, dataset_name="tourism_yearly"
    ),
    "cif_2016": partial(generate_forecasting_dataset, dataset_name="cif_2016"),
    "london_smart_meters_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="london_smart_meters_without_missing",
    ),
    "wind_farms_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="wind_farms_without_missing",
    ),
    "car_parts_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="car_parts_without_missing",
    ),
    "dominick": partial(generate_forecasting_dataset, dataset_name="dominick"),
    "fred_md": partial(generate_forecasting_dataset, dataset_name="fred_md"),
    "pedestrian_counts": partial(
        generate_forecasting_dataset, dataset_name="pedestrian_counts"
    ),
    "hospital": partial(generate_forecasting_dataset, dataset_name="hospital"),
    "covid_deaths": partial(
        generate_forecasting_dataset, dataset_name="covid_deaths"
    ),
    "kdd_cup_2018_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="kdd_cup_2018_without_missing",
    ),
    "weather": partial(generate_forecasting_dataset, dataset_name="weather"),
    "m3_monthly": partial(generate_m3_dataset, m3_freq="monthly"),
    "m3_quarterly": partial(generate_m3_dataset, m3_freq="quarterly"),
    "m3_yearly": partial(generate_m3_dataset, m3_freq="yearly"),
    "m3_other": partial(generate_m3_dataset, m3_freq="other"),
    "m4_hourly": partial(
        generate_m4_dataset,
        m4_freq="Hourly",
        pandas_freq="H",
        prediction_length=48,
    ),
    "m4_daily": partial(
        generate_m4_dataset,
        m4_freq="Daily",
        pandas_freq="D",
        prediction_length=14,
    ),
    "m4_weekly": partial(
        generate_m4_dataset,
        m4_freq="Weekly",
        pandas_freq="W",
        prediction_length=13,
    ),
    "m4_monthly": partial(
        generate_m4_dataset,
        m4_freq="Monthly",
        pandas_freq="M",
        prediction_length=18,
    ),
    "m4_quarterly": partial(
        generate_m4_dataset,
        m4_freq="Quarterly",
        pandas_freq="Q",
        prediction_length=8,
    ),
    "m4_yearly": partial(
        generate_m4_dataset,
        m4_freq="Yearly",
        pandas_freq="Y",
        prediction_length=6,
    ),
    "m5": partial(
        generate_m5_dataset,
        pandas_freq="D",
        prediction_length=28,
        m5_file_path=get_download_path() / "m5",
    ),
    "uber_tlc_daily": partial(
        generate_uber_dataset, uber_freq="Daily", prediction_length=7
    ),
    "uber_tlc_hourly": partial(
        generate_uber_dataset, uber_freq="Hourly", prediction_length=24
    ),
    "airpassengers": partial(generate_airpassengers_dataset),
    "australian_electricity_demand": partial(
        generate_forecasting_dataset,
        dataset_name="australian_electricity_demand",
    ),
    "electricity_hourly": partial(
        generate_forecasting_dataset,
        dataset_name="electricity_hourly",
    ),
    "electricity_weekly": partial(
        generate_forecasting_dataset,
        dataset_name="electricity_weekly",
    ),
    "rideshare_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="rideshare_without_missing",
    ),
    "saugeenday": partial(
        generate_forecasting_dataset,
        dataset_name="saugeenday",
    ),
    "solar_10_minutes": partial(
        generate_forecasting_dataset,
        dataset_name="solar_10_minutes",
    ),
    "solar_weekly": partial(
        generate_forecasting_dataset,
        dataset_name="solar_weekly",
    ),
    "sunspot_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="sunspot_without_missing",
    ),
    "temperature_rain_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="temperature_rain_without_missing",
    ),
    "vehicle_trips_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="vehicle_trips_without_missing",
    ),
}


dataset_names = list(dataset_recipes.keys())

default_dataset_path = get_download_path() / "datasets"

default_dataset_writer = JsonLinesWriter()


def materialize_dataset(
    dataset_name: str,
    path: Path = default_dataset_path,
    regenerate: bool = False,
    dataset_writer: DatasetWriter = default_dataset_writer,
    prediction_length: Optional[int] = None,
) -> Path:
    """
    Ensures that the dataset is materialized under the `path / dataset_name`
    path.

    Parameters
    ----------
    dataset_name
        Name of the dataset, for instance "m4_hourly".
    regenerate
        Whether to regenerate the dataset even if a local file is present.
        If this flag is False and the file is present, the dataset will not
        be downloaded again.
    path
        Where the dataset should be saved.
    prediction_length
        The prediction length to be used for the dataset. If None, the default
        prediction length will be used. The prediction length might not be
        available for all datasets.

    Returns
    -------
        The path where the dataset is materialized
    """
    assert dataset_name in dataset_recipes.keys(), (
        f"{dataset_name} is not present, please choose one from "
        f"{sorted(dataset_recipes)}."
    )

    path.mkdir(parents=True, exist_ok=True)
    dataset_path = path / dataset_name

    dataset_recipe = dataset_recipes[dataset_name]

    if not dataset_path.exists() or regenerate:
        logging.info(f"downloading and processing {dataset_name}")
        if dataset_path.exists():
            # If regenerating, we need to remove the directory contents
            shutil.rmtree(dataset_path)
            dataset_path.mkdir()

        # Optionally pass prediction length to not override any non-None
        # defaults (e.g. for M4)
        kwargs: Dict[str, Any] = {"dataset_path": dataset_path}
        if prediction_length is not None:
            kwargs["prediction_length"] = prediction_length
        kwargs["dataset_writer"] = dataset_writer
        dataset_recipe(**kwargs)
    else:
        logging.info(
            f"using dataset already processed in path {dataset_path}."
        )

    return dataset_path


def get_dataset(
    dataset_name: str,
    path: Path = default_dataset_path,
    regenerate: bool = False,
    dataset_writer: DatasetWriter = default_dataset_writer,
    prediction_length: Optional[int] = None,
) -> TrainDatasets:
    """
    Get a repository dataset.

    The datasets that can be obtained through this function have been used
    with different processing over time by several papers (e.g., [SFG17]_,
    [LCY+18]_, and [YRD15]_) or are obtained through the `Monash Time Series
    Forecasting Repository <https://forecastingdata.org/>`_.

    Parameters
    ----------
    dataset_name
        Name of the dataset, for instance "m4_hourly".
    regenerate
        Whether to regenerate the dataset even if a local file is present.
        If this flag is False and the file is present, the dataset will not
        be downloaded again.
    path
        Where the dataset should be saved.
    prediction_length
        The prediction length to be used for the dataset. If None, the default
        prediction length will be used. If the dataset is already materialized,
        setting this option to a different value does not have an effect.
        Make sure to set `regenerate=True` in this case. Note that some
        datasets from the Monash Time Series Forecasting Repository do not
        actually have a default prediction length -- the default then depends
        on the frequency of the data:
        - Minutely data --> prediction length of 60 (one hour)
        - Hourly data --> prediction length of 48 (two days)
        - Daily data --> prediction length of 30 (one month)
        - Weekly data --> prediction length of 8 (two months)
        - Monthly data --> prediction length of 12 (one year)
        - Yearly data --> prediction length of 4 (four years)

    Returns
    -------
        Dataset obtained by either downloading or reloading from local file.
    """
    dataset_path = materialize_dataset(
        dataset_name,
        path,
        regenerate,
        dataset_writer,
        prediction_length,
    )

    return load_datasets(
        metadata=dataset_path,
        train=dataset_path / "train",
        test=dataset_path / "test",
    )


if __name__ == "__main__":
    for dataset in dataset_names:
        print(f"generate {dataset}")
        ds = get_dataset(dataset, regenerate=True)
        print(ds.metadata)
        print(len(ds.train))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .representation import Representation


class MeanScaling(Representation):
    """
    A class representing a mean scaling approach. Inputs are simply rescaled
    based on their mean.

    Parameters
    ----------
    minimum_scale
        The minimum value to which re-scaled values will be clipped to.
        (default: 1e-10)
    clip_max
        The maximum value to which re-scaled values will be clipped to.
        Negative values will be clipped at -clip_max and positive values at
        clip_max. (default: None)
    """

    @validated()
    def __init__(
        self,
        scale_min: float = 1e-10,
        clip_max: Optional[float] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.scale_min = scale_min
        self.clip_max = clip_max

    def compute_scale(
        self, F, data: Tensor, observed_indicator: Tensor  # shapes (N, T, C)
    ) -> Tensor:
        # these will have shape (N, C)
        num_observed = F.sum(observed_indicator, axis=1)
        sum_observed = (data.abs() * observed_indicator).sum(axis=1)

        # first compute a global scale per-dimension
        total_observed = num_observed.sum(axis=0)
        denominator = F.maximum(total_observed, 1.0)
        default_scale = sum_observed.sum(axis=0) / denominator  # shape (C, )

        # then compute a per-item, per-dimension scale
        denominator = F.maximum(num_observed, 1.0)
        scale = sum_observed / denominator  # shape (N, C)

        # use per-batch scale when no element is observed
        # or when the sequence contains only zeros
        cond = F.broadcast_greater(sum_observed, F.zeros_like(sum_observed))
        scale = F.where(
            cond,
            scale,
            F.broadcast_mul(default_scale, F.ones_like(num_observed)),
        )

        return F.maximum(scale, self.scale_min)

    def hybrid_forward(
        self,
        F,
        data: Tensor,
        observed_indicator: Tensor,
        scale: Optional[Tensor],
        rep_params: List[Tensor],
        **kwargs,
    ) -> Tuple[Tensor, Tensor, List[Tensor]]:
        data = F.cast(data, dtype="float32")

        if scale is None:
            scale = self.compute_scale(F, data, observed_indicator)
            scale = scale.expand_dims(axis=1)

        scaled_data = F.broadcast_div(data, scale)

        if self.clip_max is not None:
            scaled_data = F.clip(scaled_data, -self.clip_max, self.clip_max)

        return scaled_data, scale, []

    def post_transform(
        self, F, samples: Tensor, scale: Tensor, rep_params: List[Tensor]
    ) -> Tensor:
        transf_samples = F.broadcast_mul(samples, scale)
        return transf_samples
